
from epyk.core.py import OrderedSet


class Plotly(object):

  @staticmethod
  def surface(data, y_columns, x_axis, z_axis):
      """

      :param data:
      :param y_columns:
      :param x_axis:
      :param z_axis:
      """
      z_a, x_a, agg_y = set(), set(), {}
      for rec in data:
        if z_axis in rec:
          z_a.add(rec[z_axis])
        if x_axis in rec:
          x_a.add(rec[x_axis])
        if z_axis in rec and x_axis in rec:
          agg_key = (rec[x_axis], rec[z_axis])
          for y in y_columns:
            agg_y.setdefault(agg_key, {})[y] = agg_y.get(agg_key, {}).get(y, 0) + float(rec[y] if rec[y] else 0)
      z_array = sorted(list(z_a))
      x_array = sorted(list(x_a))
      naps = {'datasets': [], 'series': [], 'python': True}
      for y in y_columns:
        nap = []
        for z in z_array:
          row = [agg_y.get((x, z), {}).get(y, 0) for x in x_array]
          nap.append(row)
        naps['datasets'].append(nap)
        naps['series'].append(y)
      return naps

  @staticmethod
  def map(data):
    return {'datasets': data, 'series': [], 'python': True}

  @staticmethod
  def countries(data, country_col, size_col, scale=False):
    aggregated = {}
    for rec in data:
      if country_col in rec:
        try:
          aggregated[rec[country_col]] = aggregated.get(rec[country_col], 0) + float(rec.get(size_col, 0))
        except: pass

    records = []
    if aggregated:
      max_value = max(aggregated.values())
      factor = scale if scale else 50 / max_value
      record = {'locations': [], 'marker': {'size': []}, 'python': True}
      for k, v in aggregated.items():
        record['locations'].append(k)
        record['marker']['size'].append(v * factor)
      records.append(record)
    return records

  @staticmethod
  def choropleth(data, country_col, size_col, scale=False):
    aggregated = {}
    for rec in data:
      if country_col in rec:
        try:
          aggregated[rec[country_col]] = aggregated.get(rec[country_col], 0) + float(rec.get(size_col, 0))
        except Exception as err:
          pass

    records = []
    if aggregated:
      max_value = max(aggregated.values())
      factor = scale if scale else 50 / max_value
      record = {'locations': [], 'z': []}
      for k, v in aggregated.items():
        record['locations'].append(k)
        record['z'].append(v * factor)
      records.append(record)
    return records

  @staticmethod
  def locations(data, long_col, lat_col, size_col, scale=False):
    aggregated = {}
    for rec in data:
      if long_col in rec and lat_col in rec:
        point = (rec[long_col], rec[lat_col])
        try:
          aggregated[point] = aggregated.get(point, 0) + float(rec.get(size_col, 0))
        except:
          pass

    records = []
    if aggregated:
      max_value = max(aggregated.values())
      factor = 1 / scale if scale else 50 / max_value
      record = {'lon': [], 'lat': [], 'marker': {'size': []}}
      for k, v in aggregated.items():
        record['lon'].append(float(k[0]))
        record['lat'].append(float(k[1]))
        record['marker']['size'].append(v * factor)
      records.append(record)
    return records

  @staticmethod
  def xy(data, y_columns, x_axis):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    """
    if data is None:
      return {'datasets': [], 'python': True, 'series': y_columns}

    agg_data = {}
    for rec in data:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    data = []
    for c in y_columns:
      series = {'x': [], 'y': []}
      for x, y in agg_data.get(c, {}).items():
        series['x'].append(x)
        series['y'].append(y)
      data.append(series)
    return {'datasets': data, 'python': True, 'series': y_columns}

  @staticmethod
  def xy_text(data, y_columns, x_axis, text=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param text: String. The column corresponding to the key in the dictionaries in the record
    """
    if text is None:
      return Plotly.xy(data, y_columns, x_axis)

    agg_data, texts = {}, {}
    for rec in data:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
          texts.setdefault(y, {})[rec[x_axis]] = rec[text]
    data = []
    for c in y_columns:
      series = {'x': [], 'y': [], 'text': []}
      for x, y in agg_data.get(c, {}).items():
        series['x'].append(x)
        series['y'].append(y)
        series['text'].append(texts.get(c, {}).get(x, ''))
      data.append(series)
    return {'datasets': data, 'python': True, 'series': y_columns}

  @staticmethod
  def xyz(data, y_columns, x_axis, z_axis):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param z_axis:
    """
    agg_data, agg_z = {}, {}
    for rec in data:
      for i, y in enumerate(y_columns):
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(
            rec[y] if rec[y] else 0)
        if z_axis is not None and i < len(z_axis):
          z_col = sum([float(rec[z]) for z in z_axis]) if isinstance(z_axis, list) else float(
            rec[z_axis] if rec[z_axis] else 0)
          agg_z.setdefault(y, {})[rec[x_axis]] = agg_z.get(y, {}).get(rec[x_axis], 0) + z_col
    labels, data = OrderedSet(), []
    for c in y_columns:
      series = {"x": [], "y": [], "z": []}
      for x, y in agg_data[c].items():
        labels.add(x)
        series['x'].append(x)
        series['y'].append(y)
        series['z'].append(agg_z.get(c, {}).get(x, 0))
      data.append(series)
    is_data = {"labels": labels, 'datasets': [], 'series': [], 'python': True}
    for i, l in enumerate(y_columns):
      is_data["datasets"].append(data[i])
      is_data["series"].append(l)
    return is_data

  @staticmethod
  def x_yz(data, y_columns, x_axis, z_axis, dy=0, dx=0, dz=0):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param z_axis:
    :param dy:
    :param dx:
    :param dz:
    """
    agg_data, agg_z = {}, {}
    for rec in data:
      for i, y in enumerate(y_columns):
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(
            rec[y] if rec[y] else 0)
        if z_axis is not None and i < len(z_axis):
          z_col = sum([float(rec.get(z, 0)) for z in z_axis]) if isinstance(z_axis, list) else float(
            rec[z_axis] if rec.get(z_axis, 0) else 0)
          agg_z.setdefault(y, {})[rec.get(x_axis, 0)] = agg_z.get(y, {}).get(rec.get(x_axis, 0), 0) + z_col
    labels, data = OrderedSet(), []
    for c in y_columns:
      series = {"x": [], "y": [], "z": []}
      for x, y in agg_data.get(c, {}).items():
        labels.add(x)
        z = agg_z.get(c, {}).get(x, 0)
        series['x'].append([x, float(x) + dx if x else 1])
        series['y'].append([y, y + dy])
        series['z'].append([z, z + dz])
      data.append(series)
    is_data = {"labels": labels, 'datasets': [], 'series': [], 'python': True}
    for i, l in enumerate(y_columns):
      is_data["datasets"].append(data[i])
      is_data["series"].append(l)
    return is_data

  @staticmethod
  def table(data, columns, dflt=''):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data: List of dict. The Python recordset
    :param columns: List. The key in the recordset to be used to build the row
    :param dflt: Optional. The default value if key is missing
    """
    result = {'values': [], 'python': True, 'header': [[c] for c in columns]}
    if data is None:
      return result

    for rec in data:
      result['values'].append([rec.get(c, dflt) for c in columns])
    return result


class Vis(object):

  @staticmethod
  def xyz(data, y_columns, x_axis, z_axis):
    """

    :param data:
    :param y_columns:
    :param x_axis:
    :param z_axis:
    """
    agg_data = {}
    for rec in data:
      key_point = (rec[x_axis], rec[z_axis])
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[key_point] = agg_data.get(y, {}).get(key_point, 0) + float(rec[y])
    labels, data = set(), []
    for i, c in enumerate(y_columns):
      series = []
      for point, y in agg_data[c].items():
        series.append({"x": float(point[0]), "y": y, 'z': float(point[1]), 'group': i})
      data.append(series)
    return data

  @staticmethod
  def xy(data, y_columns, x_axis):
    """

    :param data:
    :param y_columns:
    :param x_axis:
    """
    agg_data = {}
    for rec in data:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    labels, data = set(), []
    for i, c in enumerate(y_columns):
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append({"x": float(x), "y": y, 'group': i})
      data.append(series)
    return data

  @staticmethod
  def timeline(data, start, content, end=None, type=None, group=None, options=None):
    """
    Description:
    -----------
    Data transformation for the Vis Timeline chart

    Attributes:
    ----------
    :param data:
    :param start: String: The column in the record for the start date
    :param content: String:
    :param end: String: Optional. The column in the record for the end date
    :param type: String. Optional.
    :param group:
    :param options:
    """
    is_data = {'datasets': [], 'python': True}
    if data is None:
      return is_data

    options = options or {}
    for rec in data:
      row = {"start": rec[start], 'content': rec.get(content, '')}
      if end is not None and end in rec:
        row['end'] = rec[end]
      if type is not None and type in rec:
        row['type'] = rec[type]
      elif 'type' in options:
        row['type'] = options['type']
      if group is not None and group in rec:
        row['group'] = rec[group]
      is_data['datasets'].append(row)
    return is_data


class ChartJs(object):

  @staticmethod
  def y(data, y_columns, x_axis):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    """
    is_data = {"labels": [], 'datasets': [], 'series': [], 'python': True}
    if data is None or y_columns is None:
      return is_data

    agg_data = {}
    for rec in data:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    labels, data = OrderedSet(), []
    for c in y_columns:
      for x, y in agg_data.get(c, {}).items():
        labels.add(x)
    is_data["labels"] = labels
    for i, y in enumerate(y_columns):
      is_data["datasets"].append([agg_data.get(y, {}).get(x) for x in labels])
      is_data["series"].append(y)
    return is_data

  @staticmethod
  def xy(data, y_columns, x_axis):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    """
    agg_data = {}
    for rec in data:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append({"x": x, "y": y})
      data.append(series)
    is_data = {"labels": [], 'datasets': [], 'series': [], 'python': True}
    for i, l in enumerate(y_columns):
      is_data["labels"].append(l)
      is_data["datasets"].append(data[i])
      is_data["series"].append(l)
    return is_data

  @staticmethod
  def xyz(data, y_columns, x_axis, z_axis=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param z_axis:
    """
    is_data = {"labels": OrderedSet(), 'datasets': [], 'series': [], 'python': True}
    if y_columns is None:
      return is_data

    agg_data, agg_r = {}, {}
    for rec in data:
      for i, y in enumerate(y_columns):
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
        if z_axis is not None and i < len(z_axis):
          agg_r.setdefault(y, {})[rec[x_axis]] = agg_r.get(y, {}).get(rec[x_axis], 0) + float(rec[z_axis[i]])
    data = []
    for c in y_columns:
      series = []
      for x, y in agg_data.get(c, {}).items():
        is_data["labels"].add(x)
        series.append({"x": x, "y": y, 'r': agg_r.get(c, {}).get(x, 1)})
      data.append(series)
    for i, l in enumerate(y_columns):
      is_data["datasets"].append(data[i])
      is_data["series"].append(l)
    return is_data


class C3(object):

  @staticmethod
  def y(data, y_columns, x_axis):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    """
    agg_data = {}
    for rec in data:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    labels, data = OrderedSet(), []
    for c in y_columns:
      for x, y in agg_data.get(c, {}).items():
        labels.add(x)
    is_data = {"labels": labels, 'datasets': [], 'series': [], 'python': True}
    for i, y in enumerate(y_columns):
      is_data["datasets"].append([agg_data.get(y, {}).get(x) for x in labels])
      is_data["series"].append(y)
    return is_data


class NVD3(object):

  @staticmethod
  def xy(data, y_columns, x_axis):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    """
    agg_data = {}
    for rec in data:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data.get(c, {}).items():
        labels.add(x)
        series.append({"x": x, "y": y})
      data.append(series)
    is_data = {"labels": [], 'datasets': [], 'series': [], 'python': True}
    for i, l in enumerate(y_columns):
      is_data["labels"].append(l)
      is_data["datasets"].append(data[i])
      is_data["series"].append(l)
    return is_data

  @staticmethod
  def labely(data, y_columns, x_axis):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    """
    agg_data = {}
    for rec in data:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data.get(c, {}).items():
        labels.add(x)
        series.append({"label": x, "y": y})
      data.append(series)
    is_data = {"labels": [], 'datasets': [], 'series': [], 'python': True}
    for i, l in enumerate(y_columns):
      is_data["labels"].append(l)
      is_data["datasets"].append(data[i])
      is_data["series"].append(l)
    return is_data


class Datatable(object):

  @staticmethod
  def table(data, columns, dflt=''):
    """
    Description:
    ------------
    Tranform the data in a list of list for Datatable

    Attributes:
    ----------
    :param data: List of dict. The Python recordset
    :param columns: List. The key in the recordset to be used to build the row
    :param dflt: Optional. The default value if key is missing
    """
    records = []
    for rec in data:
      records.append([rec.get(c, dflt) for c in columns])
    return records


class Google(object):

  @staticmethod
  def y(data, y_columns, x_axis):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    """
    is_data = {"labels": [], 'datasets': [], 'series': [], 'python': True}
    if data is None:
      return is_data

    agg_data = {}
    for rec in data:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    labels, data = OrderedSet(), []
    for c in y_columns:
      for x, y in agg_data.get(c, {}).items():
        labels.add(x)
    is_data["labels"] = labels
    for x in labels:
      is_data["datasets"].append([x] + [agg_data.get(y, {}).get(x) for y in y_columns])
    is_data["series"] = y_columns
    is_data["x"] = x_axis
    return is_data

  @staticmethod
  def table(data, rows, cols):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data:
    :param rows:
    :param cols:
    """
    is_data = {"rows": rows, 'datasets': [], 'cols': cols, 'python': True}

    for rec in data:
      is_data['datasets'].append([rec.get(c, '') for c in rows + cols])

    return is_data



# <span class="fas fa-spinner fa-spin" style="padding:auto;margin:auto;font-size:65px;"></span>

STATIC_PAGE = '''
<!DOCTYPE html>
<html lang="en" style="height:100%%">
<head>
%(header)s
%(jsImports)s
%(cssImports)s
<style>
%(cssStyle)s
</style>
</head>
%(body)s
<script>
%(jsFrgs)s
</script>
</html>
'''

# %(jsFrgs_in_req)s
# %(jsImports)s
JUPYTER = '''
<head>
<style>
    %(cssStyle)s
</style>
%(cssImports)s
</head>
<body>
%(content)s
<script>
require.config({ paths:%(paths)s});
%(jsFrgs_in_req)s
</script>
<body>
'''


JUPYTERLAB = '''
<head>
<style>
    %(cssStyle)s
</style>
%(cssImports)s
%(jsImports)s
</head>
<body>
%(content)s
<script>
%(jsFrgs)s
</script>
<body>
'''



"""
Module dedicated to wrap the Javascript Date

Related Pages:

		https://www.w3schools.com/jsref/jsref_obj_date.asp
"""


import json

from epyk.core.js.primitives import JsObject
from epyk.core.js import JsUtils

MONTHS = ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December']


class JsDate(JsObject.JsObject):
  _jsClass = "Date"

  def __init__(self, data=None, varName=None, setVar=False, isPyData=False):
    """
    Description:
    -----------

    Related Pages:

			https://www.w3schools.com/jsref/jsref_obj_date.asp

    Attributes:
    ----------
    :param data: Required,
    :param varName: Optional,
    :param setVar: Optional,
    """
    if setVar:
      if data is not None:
        isPyData = False
        data = "new Date(%s)" % json.dumps(data) if isPyData else "new Date(%s)" % data
    if data is None:
      isPyData = False
      data = "new Date()"
    if not hasattr(data, 'varName') and isPyData:
      isPyData = True
      data = "new Date(%s)" % json.dumps(data)
    super(JsDate, self).__init__(data=data, varName=varName, setVar=setVar, isPyData=isPyData)

  @classmethod
  def get(cls, varName):
    """
    Description:
    -----------
    Get the Javascript Object by its reference

    Usage::

      JsDate.new("2019-05-03", varName="MyDate")
      JsDate.get("MyDate")

    Related Pages:

			https://www.w3schools.com/jsref/jsref_obj_date.asp

    Attributes:
    ----------
    :param varName: The Javascript object reference

    :return: The python Javascript object
    """
    if "-" in varName:
      # assume it a valid date
      return cls(data=None, varName='new Date("%s")' % varName, setVar=False)

    JsUtils.getJsValid(varName)
    return cls(data=None, varName=varName, setVar=False)

  @property
  def isWeedend(self):
    """
    Description:
    -----------

    Usage::

      jsObj.objects.date.get("dateTest").isWeedend
    """
    from epyk.core.js.primitives import JsBoolean
    return JsBoolean.JsBoolean("(%(varId)s.getDay() === 6) || (%(varId)s.getDay() === 0)" % {"varId": self.varId}, isPyData=False)

  @staticmethod
  def now():
    """
    Description:
    -----------
    The Date.now() method returns the number of milliseconds since January 1, 1970 00:00:00 UTC.

    Usage::

      jsObj.objects.date.now()

    Related Pages:

			https://www.w3schools.com/jsref/jsref_now.asp

    :return: A Number, representing the number of milliseconds since midnight January 1, 1970
    """
    from epyk.core.js.primitives import JsNumber

    return JsNumber.JsNumber("Date.now()", isPyData=False)

  @staticmethod
  def today():
    """
    Description:
    -----------
    Return the String date in the standard format YYYY-MM-DD

    Usage::

      rptObj.js.objects.date.today()

    Related Pages:

			https://www.w3schools.com/jsref/jsref_toisostring.asp

    :return: A Python / Javascript object
    """
    from epyk.core.js.primitives import JsString
    return JsString.JsString("function(){return new Date}().toISOString().slice(0, 10)", isPyData=False)

  def getDate(self):
    """
    Description:
    -----------
    The getDate() method returns the day of the month (from 1 to 31) for the specified date.

    Related Pages:

			https://www.w3schools.com/jsref/jsref_getdate.asp

    :return:
    """
    return JsDate("%s.getDate()" % self.varId, isPyData=False)

  def getDay(self):
    """
    Description:
    -----------
    The getDay() method returns the day of the week (from 0 to 6) for the specified date.

    Usage::

      jsObj.objects.date.new("2019-01-01", varName="dateTest")
      jsObj.objects.date.get("dateTest").getDay()

    Related Pages:

			https://www.w3schools.com/jsref/jsref_getday.asp

    :return:
    """
    from epyk.core.js.primitives import JsNumber
    return JsNumber.JsNumber("%s.getDay()" % self.varId, isPyData=False)

  def getFullYear(self):
    """
    Description:
    -----------
    The getFullYear() method returns the year (four digits for dates between year 1000 and 9999) of the specified date.

    Usage::

      jsObj.objects.date.new("2019-01-01", varName="dateTest")
      jsObj.objects.date.get("dateTest").getFullYear()

    Related Pages:

			https://www.w3schools.com/jsref/jsref_getfullyear.asp

    :return: A Number, representing the year of the specified date
    """
    from epyk.core.js.primitives import JsNumber

    return JsNumber.JsNumber("%s.getFullYear()" % self.varId, isPyData=False)

  def getHours(self):
    """
    Description:
    -----------
    The getHours() method returns the hour (from 0 to 23) of the specified date and time.

    Usage::

      jsObj.objects.date.new("2019-01-01", varName="dateTest")
      jsObj.objects.date.get("dateTest").getHours()

    Related Pages:

			https://www.w3schools.com/jsref/jsref_gethours.asp

    :return: A Number, from 0 to 23, representing the hour
    """
    from epyk.core.js.primitives import JsNumber

    return JsNumber.JsNumber("%s.getHours()" % self.varId, isPyData=False)

  def getMilliseconds(self):
    """
    Description:
    -----------
    The getMilliseconds() method returns the milliseconds (from 0 to 999) of the specified date and time.

    Usage::

      jsObj.objects.date.new("2019-01-01", varName="dateTest")
      jsObj.objects.date.get("dateTest").getMilliseconds()

    Related Pages:

			https://www.w3schools.com/jsref/jsref_getmilliseconds.asp

    :return: A Number, from 0 to 999, representing milliseconds
    """
    from epyk.core.js.primitives import JsNumber

    return JsNumber.JsNumber("%s.getMilliseconds()" % self.varId, isPyData=False)

  def getMonth(self):
    """
    Description:
    -----------
    The getMonth() method returns the month (from 0 to 11) for the specified date, according to local time.

    Usage::

      jsObj.objects.date.new("2019-01-01", varName="dateTest")
      jsObj.objects.date.get("dateTest").getMonth()

    Related Pages:

			https://www.w3schools.com/jsref/jsref_getmonth.asp

    :return: A Number, from 0 to 11, representing the month
    """
    from epyk.core.js.primitives import JsNumber
    return JsNumber.JsNumber("%s.getMonth()" % self.varId, isPyData=False)

  def getMonthName(self):
    """
    Description:
    -----------
    Use getMonth() method returns the month the month name from the definition in the module .

    Usage::

      jsObj.objects.date.new("2019-01-01", varName="dateTest")
      jsObj.objects.date.get("dateTest").getMonthName()

    Related Pages:

			https://www.w3schools.com/jsref/jsref_getmonth.asp

    :return: A Number, from 0 to 11, representing the month
    """
    from epyk.core.js.primitives import JsNumber
    return JsNumber.JsNumber("(function(x){return %s[x]})(%s.getMonth())" % (MONTHS, self.varId), isPyData=False)

  def setDate(self, day):
    """
    Description:
    -----------
    The setDate() method sets the day of the month to the date object.

    Usage::

      jsObj.objects.date.new("2019-01-01", varName="dateTest")
      jsObj.objects.date.get("dateTest").setDate(5)

    Related Pages:

			https://www.w3schools.com/jsref/jsref_setdate.asp

    Attributes:
    ----------
    :param day: Required. An integer representing the day of a month.

    :return: A Number, representing the number of milliseconds between the date object and midnight January 1 1970
    """
    return JsDate("%s.setDate(%s)" % (self.varId, day), isPyData=False)

  def setMonth(self, month, day=None):
    """
    Description:
    -----------
    The setMonth() method sets the month of a date object.
    Return a new date object

    Usage::

      jsObj.objects.date.new("2019-01-01", varName="dateTest")
      jsObj.objects.date.get("dateTest").setMonth(5)

    Related Pages:

			https://www.w3schools.com/jsref/jsref_setmonth.asp

    Attributes:
    ----------
    :param month: Required. An integer representing the month
    :param day: Optional. An integer representing the day of month

    :return: A Number, representing the number of milliseconds between the date object and midnight January 1 1970
    """
    if day is not None:
      return JsDate("new Date(%s.setMonth(%s, %s))" % (self.varId, month, day), isPyData=False)

    return JsDate("new Date(%s.setMonth(%s))" % (self.varId, month), isPyData=False)

  def toDateString(self):
    """
    Description:
    -----------
    The toDateString() method converts the date (not the time) of a Date object into a readable string.

    Usage::

      jsType.date.new(varName="Test")
      jsType.date.get("Test").toDateString()

    Related Pages:

			https://www.w3schools.com/jsref/jsref_todatestring.asp

    :return: A String, representing the date as a string
    """
    from epyk.core.js.primitives import JsString

    return JsString.JsString("%s.toDateString()" % self.varId, isPyData=False)

  def toISOString(self):
    """
    Description:
    -----------
    The toISOString() method converts a Date object into a string, using the ISO standard.

    Usage::

      jsType.date.new(varName="Test")
      jsType.date.get("Test").toISOString()

    Related Pages:

			https://www.w3schools.com/jsref/jsref_toisostring.asp

    :return: A String, representing the date and time using the ISO standard format
    """
    from epyk.core.js.primitives import JsString

    return JsString.JsString("%s.toISOString()" % self.varId, isPyData=False)

  def getStrDate(self):
    """
    Description:
    -----------
    Return the String date in the standard format YYYY-MM-DD

    Usage::

      jsObj.objects.date.new("2019-01-01", varName="dateTest")
      jsType.date.get("dateTest").getStrDate()

    Related Pages:

			https://www.w3schools.com/jsref/jsref_toisostring.asp

    :return: A Python / Javascript object
    """
    from epyk.core.js.primitives import JsString
    return JsString.JsString("%s.toISOString().slice(0, 10)" % self.varId, isPyData=False)

  def getStrTimeStamp(self):
    """
    Description:
    -----------
    The toISOString() method converts a Date object into a string, using the ISO standard.

    Usage::

      jsObj.objects.date.new("2019-01-01", varName="dateTest")
      jsType.date.get("dateTest").getStrTimeStamp()

    Related Pages:

			https://www.w3schools.com/jsref/jsref_toisostring.asp

    :return: A Python / Javascript object
    """
    from epyk.core.js.primitives import JsString
    return JsString.JsString("%s.toISOString().replace('T', ' ').slice(0, 19)" % self.varId, isPyData=False)

  def getTime(self, in_seconds=True):
    """
    Description:
    -----------
    To get the unix timestamp using JavaScript you need to use the getTime() function of the build in Date object.
    As this returns the number of milliseconds then we must divide the number by 1000 and round it in order to get the timestamp in seconds.

    Attributes:
    ----------
    :param in_seconds: Boolean. In second conversion of the Javascript timestamp
    """
    from epyk.core.js.primitives import JsNumber

    if in_seconds:
      return JsNumber.JsNumber("%s.getTime()/1000" % self.varId, isPyData=False)

    return JsNumber.JsNumber("%s.getTime()" % self.varId, isPyData=False)

  def add(self, n):
    """
    Description:
    -----------
    Simple wrapper to the Javascript add method.
    This will just return the Js string corresponding to the add

    This function is used in the addDays method

    Usage::

      jsObj.objects.date.this().getDate().add(1)

    Attributes:
    ----------
    :param n: Integer the number of days

    :return: A Python Js object
    """
    jsData = JsUtils.jsConvertData(n, None)
    return super(JsDate, self).add(jsData)

  def addDays(self, jsObj, n, weekend=False):
    """
    Description:
    -----------
    Add some days to a Javascript date

    Usage::

      jsObj.objects.date.get("MyDate").addDays(jsObj, 6).getStrDate()

    Related Pages:

			https://stackoverflow.com/questions/563406/add-days-to-javascript-date

    Attributes:
    ----------
    :param jsObj: The internal JS object used to store the prototype extension
    :param n: The number of days to be added
    :param weekend: Boolean flag to specify if the weekends should be considered in the count. Default False

    :type jsObj: epyk.Lib.js.Js.JsBase
    """
    jsObj.extendProto(self, "addDays", [
      jsObj.objects.date.this().setDate(jsObj.objects.date.this().getDate().add(jsObj.parseInt(jsObj.objects.get("n")))),
      jsObj.if_([jsObj.objects.date.this().isWeedend, jsObj.objects.boolean.get("weekend").not_], [
        jsObj.if_(jsObj.objects.date.this().getDay() == 0, jsObj.objects.date.this().setDate(jsObj.objects.date.this().getDate().add(1))),
        jsObj.if_(jsObj.objects.date.this().getDay() == 6, jsObj.objects.date.this().setDate(jsObj.objects.date.this().getDate().add(2)))
      ]),
      jsObj.return_(jsObj.objects.date.this())
    ], pmts=["n", "weekend"])
    return JsDate("%s.addDays(%s, %s)" % (self.varId, n, json.dumps(weekend)))




from epyk.core.html.options import Options


class OptionsButton(Options):

  @property
  def multiple(self):
    """
    Description:
    ------------
    Property to define if multiple buttons can be selected at the same time
    Default value is false
    """
    return self.get(False)

  @multiple.setter
  def multiple(self, bool):
    self.set(bool)

  @property
  def group(self):
    """
    Description:
    ------------
    Property to set the group name of a button
    """
    return self._report.attr.get('name')

  @group.setter
  def group(self, val):
    self._report.set_attrs(name='name', value=val)


class OptionsBadge(Options):

  @property
  def badge_css(self):
    """
    Description:
    ------------
    """
    return self.get()

  @badge_css.setter
  def badge_css(self, css):
    if hasattr(self, '_report') and hasattr(self._report, 'link'):
      self._report.link.css(css)
    cssOpts = self.get({})
    cssOpts.update(css)
    self.set(cssOpts)

  @property
  def badge_position(self):
    """
    Description:
    ------------

    :param position:
    """
    return self.get()

  @badge_position.setter
  def badge_position(self, position):
    if position == 'left':
      self.set({"position": 'relative'}, name='badge_css')
    else:
      self.set({"position": 'relative', "top": "-4px", "right": "11px"}, name='badge_css')
    self.set(position)





from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObjects
from epyk.core.js.packages import JsC3


class Billboard(JsC3.C3):

  def export(self):
    """
    Export chart as an image.

    https://naver.github.io/billboard.js/release/latest/doc/Chart.html#export
    """
    return JsObjects.JsVoid("%s.export()" % self._selector)

  def revert(self, target=None):
    """
    Export chart as an image.

    https://naver.github.io/billboard.js/release/latest/doc/Chart.html#export

    :param target: String | Array. ids to be reverted
    """
    target = JsUtils.jsConvertData(target, None)
    return JsObjects.JsVoid("%s.revert(%s)" % (self._selector, target))




from epyk.core import html


class Inputs(object):
  def __init__(self, context):
    self.context = context

  def d_text(self, text="", placeholder='', width=(100, "%"), height=(None, "px"), htmlCode=None,
            options=None, attrs=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.inputs.d_text()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.Input`

    Attributes:
    ----------
    :param text:
    :param placeholder:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param attrs:
    :param profile:
    """
    options = options or {}
    attrs = attrs or {}
    html_input = html.HtmlInput.Input(self.context.rptObj, text, placeholder, width, height, htmlCode, options, attrs, profile)
    html_input.style.css.margin_bottom = '2px'
    return html_input

  def d_radio(self, flag=False, group_name=None, placeholder='', width=(100, "%"), height=(None, "px"), htmlCode=None,
            options=None, attrs=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.inputs.d_radio()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.InputRadio`

    Attributes:
    ----------
    :param flag:
    :param placeholder:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param attrs:
    :param profile:
    """
    options = options or {}
    attrs = attrs or {}
    html_input = html.HtmlInput.InputRadio(self.context.rptObj, flag, group_name, placeholder, width, height, htmlCode,
                                           options, attrs, profile)
    return html_input

  def d_search(self, text="", placeholder='', width=(100, "%"), height=(None, "px"), htmlCode=None, options=None, attrs=None, profile=None):
    """
    Description:
    ------------
    One of the new types of inputs in HTML5 is search

    Usage::

      rptObj.ui.inputs.d_search("")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.Input`

    Related Pages:

			https://developer.mozilla.org/fr/docs/Web/HTML/Element/Input/search
    https://css-tricks.com/webkit-html5-search-inputs/

    Attributes:
    ----------
    :param text:
    :param placeholder:
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height:
    :param htmlCode:
    :param options:
    :param attrs:
    :param profile:
    """
    attrs = attrs or {}
    html_search = html.HtmlInput.Input(self.context.rptObj, text, placeholder, width, height, htmlCode,
                                       options, attrs, profile)
    attrs.update({"type": 'search'})
    return html_search

  def password(self, text="", placeholder='', width=(100, "%"), height=(None, "px"), htmlCode=None, options=None, attrs=None, profile=None):
    """

    Description:
    ------------
    Input field that will hide characters typed in

    Usage::

      rptObj.ui.inputs.password(placeholder="Password")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.Input`

    :param text:
    :param placeholder:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param attrs:
    :param profile:
    :return:
    """
    attrs = attrs or {}
    attrs.update({"type": 'password'})
    return html.HtmlInput.Input(self.context.rptObj, text, placeholder, width, height, htmlCode, options, attrs, profile)

  def file(self, text="", placeholder='', width=(100, "%"), height=(None, "px"), htmlCode=None, options=None, attrs=None, profile=None):
    """
    Input file object.

    Description:
    ------------
    Input field that will hide characters typed in

    Usage::

      rptObj.ui.inputs.file()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.File`

    :param text:
    :param placeholder:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param attrs:
    :param profile:
    :return:
    """
    attrs = attrs or {}
    attrs.update({"type": 'file'})
    return html.HtmlInput.Input(self.context.rptObj, text, placeholder, width, height, htmlCode, options, attrs, profile)

  def d_time(self, text="", placeholder='', width=(139, "px"), height=(None, "px"), htmlCode=None, options=None, attrs=None, profile=None):
    """
    Description:
    ------------

    Usage::

      date = rptObj.ui.inputs.d_time()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.InputTime`

    Attributes:
    ----------
    :param text:
    :param placeholder:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param attrs:
    :param profile:
    """
    dflt_options = {'timeFormat': 'HH:mm:ss'}
    dflt_options.update(options or {})
    html_input_t = html.HtmlInput.InputTime(self.context.rptObj, text, placeholder, width, height, htmlCode, dflt_options, attrs or {}, profile)
    return html_input_t

  def d_date(self, text, placeholder='', width=(140, "px"), height=(None, "px"), htmlCode=None, options=None, attrs=None, profile=None):
    """
    Description:
    ------------

    Usage::

      date = rptObj.ui.inputs.d_date()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.InputDate`

    Attributes:
    ----------
    :param text:
    :param placeholder:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param attrs:
    :param profile:
    """
    html_date = html.HtmlInput.InputDate(self.context.rptObj, text, placeholder, width, height, htmlCode, options, attrs or {}, profile)
    return html_date

  def d_int(self, value="", placeholder='', width=(100, "%"), height=(None, "px"), htmlCode=None, options=None, attrs=None, profile=None):
    """

    Description:
    ------------

    Usage::

      date = rptObj.ui.inputs.d_int()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.InputInteger`

    :param value:
    :param placeholder:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param attrs:
    :param profile:
    """
    attrs = attrs or {}
    attrs.update({"type": 'number'})
    html_integer = html.HtmlInput.InputInteger(self.context.rptObj, value, placeholder, width, height, htmlCode, options, attrs, profile)
    return html_integer

  def d_range(self, value, min=0, max=100, step=1, placeholder='', width=(100, "%"), height=(None, "px"), htmlCode=None,
              options=None, attrs=None, profile=None):
    attrs = attrs or {}
    attrs.update({"type": 'range'})
    html_range = html.HtmlInput.InputRange(self.context.rptObj, value, min, max, step, placeholder, width, height,
                                           htmlCode, options or {}, attrs, profile)
    return html_range

  def _output(self, value="", options=None, profile=False):
    """
    Description:
    ------------
    Create a HTML output object

    Usage::

      rptObj.ui.inputs._output("test output")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.Output`

    Attributes:
    ----------
    :param value:
    :param options:
    :param profile:
    """
    html_output = html.HtmlInput.Output(self.context.rptObj, value)
    return html_output

  def textarea(self, text="", width=(100, '%'), rows=5, placeholder=None, background_color=None, htmlCode=None,
               options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.inputs.textarea("Test")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.TextArea`

    Related Pages:

			https://www.w3schools.com/tags/tag_textarea.asp

    Attributes:
    ----------
    :param text:
    :param width:
    :param rows:
    :param background_color:
    :param htmlCode:
    :param options:
    :param profile:
    """
    dfltOptions = {"spellcheck": False, 'selectable': False}
    dfltOptions.update(options or {})
    html_t_area = html.HtmlInput.TextArea(self.context.rptObj, text, width, rows, placeholder, background_color, htmlCode, dfltOptions, profile)
    return html_t_area

  def autocomplete(self, text="", placeholder='', width=(100, "%"), height=(None, "px"), htmlCode=None, options=None, attrs=None, profile=None):
    """
    Description:
    ------------
    Enables users to quickly find and select from a pre-populated list of values as they type, leveraging searching and filtering.

    Usage::

      rptObj.ui.inputs.autocomplete("Test")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.AutoComplete`


    Related Pages:

			https://jqueryui.com/autocomplete/


    Attributes:
    ----------
    :param text:
    :param placeholder:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param attrs:
    :param profile:
    """
    options = options or {}
    attrs = attrs or {}
    html_input = html.HtmlInput.AutoComplete(self.context.rptObj, text, placeholder, width, height, htmlCode, options, attrs, profile)
    return html_input

  def input(self, text="", placeholder='', width=(100, "%"), height=(None, "px"), htmlCode=None, options=None, attrs=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.inputs.input("Test")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.Input`

    Attributes:
    ----------
    :param text:
    :param placeholder:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param attrs:
    :param profile:

    :rtype: html.HtmlInput.Input
    :return:
    """
    return self.d_text(text, placeholder, width, height, htmlCode, options, attrs, profile)

  def hidden(self, text="", placeholder='', width=(100, "%"), height=(None, "px"), htmlCode=None, options=None, attrs=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.inputs.hidden("Test")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.Input`

    Attributes:
    ----------
    :param text:
    :param placeholder:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param attrs:
    :param profile:

    :rtype: html.HtmlInput.Input
    """
    input = self.d_text(text, placeholder, width, height, htmlCode, options, attrs, profile)
    input.style.css.display = None
    return input

  def checkbox(self, flag, label="", group_name=None, width=(None, "%"), height=(None, "px"), htmlCode=None, options=None, attrs=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.inputs.checkbox(False)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.Checkbox`

    Attributes:
    ----------
    :param flag:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param attrs:
    :param profile:
    """
    options = options or {}
    attrs = attrs or {}
    html_coech = html.HtmlInput.Checkbox(self.context.rptObj, flag, label, group_name, width, height, htmlCode,
                                         options, attrs, profile)
    return html_coech

  def radio(self, flag, label=None, group_name=None, icon=None, width=(None, "%"), height=(None, "px"), htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.inputs.radio(False, label="radio")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.Radio`

    Related Pages:

			https://www.w3schools.com/tags/att_input_type_radio.asp

    Attributes:
    ----------
    :param flag:
    :param label:
    :param group_name:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    html_radio = html.HtmlInput.Radio(self.context.rptObj, flag, label, group_name, icon, width, height, htmlCode,
                                      helper, options, profile)
    return html_radio

  def editor(self, text="", language='python', width=(100, "%"), height=(None, "px"), htmlCode=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.inputs.editor()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTextEditor.Editor`

    Attributes:
    ----------
    :param text:
    :param language:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param profile:
    """
    dflt_options = {"lineNumbers": True, 'mode': 'css', 'matchBrackets': True, 'styleActiveLine': True, 'autoRefresh': True}
    if options is not None:
      dflt_options.update(options)
    editor = html.HtmlTextEditor.Editor(self.context.rptObj, text, language, width, height, htmlCode, dflt_options, profile)
    return editor

  def cell(self, text=None, language='python', width=(100, "%"), height=(100, "px"), htmlCode=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.inputs.cell()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTextEditor.Cell`

    Attributes:
    ----------
    :param text:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    dflt_options = {"lineNumbers": True, 'mode': 'css', 'matchBrackets': True, 'styleActiveLine': True, 'autoRefresh': True}
    if options is not None:
      dflt_options.update(options)
    html_cell = html.HtmlTextEditor.Cell(self.context.rptObj, text, language, width, height, htmlCode, dflt_options, profile)
    return html_cell

  def search(self, text='', placeholder='Search..', color=None, height=(None, "px"), htmlCode=None,
             tooltip='', extensible=False, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.inputs.search()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTextEditor.Cell`

    Related Pages:

			https://www.w3schools.com/howto/tryit.asp?filename=tryhow_css_anim_search

    Attributes:
    ----------
    :param text:
    :param placeholder:
    :param color:
    :param height:
    :param htmlCode:
    :param tooltip:
    :param extensible:
    :param profile:
    """
    html_s = html.HtmlInput.Search(self.context.rptObj, text, placeholder, color, height, htmlCode, tooltip,
                                   extensible, profile)
    return html_s

  def label(self, label, text="", placeholder='', width=(100, "%"), height=(None, "px"), htmlCode=None,
            options=None, attrs=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.inputs.label()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.Input`
      - :class:`epyk.core.html.HtmlText.Label`
      - :class:`epyk.core.html.HtmlContainer.Div`

    Attributes:
    ----------
    :return:
    """
    label = self.context.rptObj.ui.texts.label(label).css({"display": 'block', 'text-align': 'left', 'margin-top': '10px',
                                                           "position": "absolute", "z-index": '20px', "font-size": '14px'})
    html_input = html.HtmlInput.Input(self.context.rptObj, text, placeholder, width, height, htmlCode,
                                      options or {}, attrs or {}, profile).css({"margin-top": '10px'})
    div = self.context.rptObj.ui.div([label, html_input])
    div.input = html_input
    div.label = label
    html_input.on('focus', [
      "document.getElementById('%s').animate({'marginTop': ['10px', '-8px']}, {duration: 50, easing: 'linear', iterations: 1, fill: 'both'})" % label.htmlCode,
    ])
    html_input.on('blur', [
      "document.getElementById('%s').animate({'marginTop': ['-8px', '10px']}, {duration: 1000, easing: 'linear', iterations: 1, fill: 'both'})" % label.htmlCode,
    ])
    return div




from epyk.core.css.styles import GrpCls
from epyk.core.css.styles.attributes import AttrClsMenu


class ClassNav(GrpCls.ClassHtml):
  @property
  def css(self):
    """
    Description:
    -----------
    Property to the underlying CSS definition to be added to the style HTML tag of a component

    :rtype: AttrClsMenu.NavBar
    """
    if self._css_struct is None:
      self._css_struct = AttrClsMenu.NavBar(self.htmlObj)
    return self._css_struct


class ClassFooter(GrpCls.ClassHtml):

  @property
  def css(self):
    """
    Description:
    -----------
    Property to the underlying CSS definition to be added to the style HTML tag of a component

    :rtype: AttrClsMenu.Footer
    """
    if self._css_struct is None:
      self._css_struct = AttrClsMenu.Footer(self.htmlObj)
    return self._css_struct


class ClassShortcut(GrpCls.ClassHtml):

  @property
  def css(self):
    """
    Description:
    -----------
    Property to the underlying CSS definition to be added to the style HTML tag of a component

    :rtype: AttrClsMenu.Footer
    """
    if self._css_struct is None:
      self._css_struct = AttrClsMenu.Footer(self.htmlObj)
    return self._css_struct





import json


class Aria(object):
  def __init__(self, htmlObj):
    self.htmlObj = htmlObj

  def set(self, arias):
    """
    Description:
    ------------
    Set multiple aria properties

    :param arias:
    """
    for k, v in arias.items():
      k = k.replace("aria-", "")
      setattr(self, k, v)

  def custom(self, key, val):
    """
    Description:
    ------------

    :param key:
    :param val:
    """
    self.htmlObj.attr["aria-%s" % key] = val

  @property
  def role(self):
    """
    Description:
    ------------

    Example of roles
    Roles: button, checkbox, menuitem, menuitemcheckbox, menuitemradio, option, radio, switch, tab or treeitem

    :return:
    """
    return self.htmlObj.attr.get("role", '')

  @role.setter
  def role(self, val):
    self.htmlObj.attr["role"] = val

  @property
  def atomic(self):
    """
    Description:
    ------------
    Indicates whether assistive technologies will present all, or only parts of, the changed region based on the change notifications defined by the aria-relevant attribute.

    Related Pages:

			https://www.w3.org/TR/wai-aria-1.1/#aria-atomic
    """
    return self.htmlObj.attr.get("aria-atomic", False)

  @atomic.setter
  def atomic(self, bool):
    self.htmlObj.attr["aria-atomic"] = bool

  @property
  def autocomplete(self):
    """
    Description:
    ------------
    Indicates whether inputting text could trigger display of one or more predictions of the user's intended value for an input and specifies how predictions would be presented if they are made.

    The aria-autocomplete property describes the type of interaction model a textbox, searchbox, or combobox employs when dynamically helping users complete text input.
    It distinguishes between two models: the inline model (aria-autocomplete="inline") that presents a value completion prediction inside the text input and the list model (aria-autocomplete="list") that presents a collection of possible values in a separate element that pops up adjacent to the text input. It is possible for an input to offer both models at the same time (aria-autocomplete="both").

    Related Pages:

			https://www.w3.org/TR/wai-aria-1.1/#aria-autocomplete
    """
    return self.htmlObj.attr.get("aria-autocomplete", False)

  @autocomplete.setter
  def autocomplete(self, bool):
    if not bool and "aria-autocomplete" in self.htmlObj.attr:
      del self.htmlObj.attr["aria-autocomplete"]
    else:
      self.htmlObj.attr["aria-autocomplete"] = None

  @property
  def busy(self):
    """
    Description:
    ------------
    Indicates an element is being modified and that assistive technologies MAY want to wait until the modifications are complete before exposing them to the user.

    The default value of aria-busy is false for all elements. When aria-busy is true for an element, assistive technologies MAY ignore changes to content owned by that element and then process all changes made during the busy period as a single, atomic update when aria-busy becomes false.

    Related Pages:

			https://www.w3.org/TR/wai-aria-1.1/#aria-busy

    :return:
    """
    return self.htmlObj.attr.get("aria-busy", False)

  @busy.setter
  def busy(self, bool):
    self.htmlObj.attr["aria-busy"] = bool

  @property
  def checked(self):
    """
    Description:
    ------------
    Indicates the current "checked" state of checkboxes, radio buttons, and other widgets. See related aria-pressed and aria-selected.

    The aria-checked attribute indicates whether the element is checked (true), unchecked (false), or represents a group of other elements that have a mixture of checked and unchecked values (mixed). Most inputs only support values of true and false, but the mixed value is supported by certain tri-state inputs such as a checkbox or menuitemcheckbox.

    Related Pages:

			https://www.w3.org/TR/wai-aria-1.1/#aria-checked

    :return:
    """
    return self.htmlObj.attr.get("aria-checked", False)

  @checked.setter
  def checked(self, bool):
    self.htmlObj.attr["aria-checked"] = bool

  @property
  def colcount(self):
    """
    Description:
    ------------
    Defines the total number of columns in a table, grid, or treegrid. See related aria-colindex.

    If all of the columns are present in the DOM, it is not necessary to set this attribute as the user agent can automatically calculate the total number of columns.
    However, if only a portion of the columns is present in the DOM at a given moment, this attribute is needed to provide an explicit indication of the number of columns in the full table.

    Related Pages:

			https://www.w3.org/TR/wai-aria-1.1/#aria-colcount
    """
    return self.htmlObj.attr.get("aria-colcount", 0)

  @colcount.setter
  def colcount(self, num):
    self.htmlObj.attr["aria-colcount"] = num

  @property
  def colindex(self):
    """
    Description:
    ------------
    Defines an element's column index or position with respect to the total number of columns within a table, grid, or treegrid. See related aria-colcount and aria-colspan.

    If all of the columns are present in the DOM, it is not necessary to set this attribute as the user agent can automatically calculate the column index of each cell or gridcell.
    However, if only a portion of the columns is present in the DOM at a given moment, this attribute is needed to provide an explicit indication of the column of each cell or gridcell with respect to the full table.

    Related Pages:

			https://www.w3.org/TR/wai-aria-1.1/#aria-colindex
    """
    return self.htmlObj.attr.get("aria-colindex", 0)

  @colindex.setter
  def colindex(self, num):
    self.htmlObj.attr["aria-colindex"] = num

  @property
  def colspan(self):
    """
    Description:
    ------------
    Defines the number of columns spanned by a cell or gridcell within a table, grid, or treegrid. See related aria-colindex and aria-rowspan.

    This attribute is intended for cells and gridcells which are not contained in a native table. When defining the column span of cells or gridcells in a native table, authors SHOULD use the host language's attribute instead of aria-colspan.
    If aria-colspan is used on an element for which the host language provides an equivalent attribute, user agents MUST ignore the value of aria-colspan and instead expose the value of the host language's attribute to assistive technologies.

    Related Pages:

			https://www.w3.org/TR/wai-aria-1.1/#aria-colspan
    """
    return self.htmlObj.attr.get("aria-colspan", 0)

  @colspan.setter
  def colspan(self, num):
    self.htmlObj.attr["aria-colspan"] = num

  @property
  def controls(self):
    """
    Description:
    ------------
    Identifies the element (or elements) whose contents or presence are controlled by the current element. See related aria-owns.

    Related Pages:

			https://www.w3.org/TR/wai-aria-1.1/#aria-controls
    """
    return self.htmlObj.attr.get("aria-controls")

  @controls.setter
  def controls(self, val):
    self.htmlObj.attr["aria-controls"] = val

  @property
  def current(self):
    """
    Description:
    ------------
    Indicates the element that represents the current item within a container or set of related elements.

    The aria-current attribute is an enumerated type. Any value not included in the list of allowed values SHOULD be treated by assistive technologies as if the value true had been provided.
    If the attribute is not present or its value is an empty string or undefined, the default value of false applies and the aria-current state MUST NOT be exposed by user agents or assistive technologies.

    Related Pages:

			https://www.w3.org/TR/wai-aria-1.1/#aria-current
    """
    return self.htmlObj.attr.get("aria-current")

  @current.setter
  def current(self, val):
    self.htmlObj.attr["aria-current"] = val

  @property
  def describedby(self):
    """
    Description:
    ------------
    Identifies the element (or elements) that describes the object. See related aria-labelledby.

    The aria-labelledby attribute is similar to the aria-describedby in that both reference other elements to calculate a text alternative, but a label should be concise, where a description is intended to provide more verbose information.

    Related Pages:

			https://www.w3.org/TR/wai-aria-1.1/#aria-describedby
    """
    return self.htmlObj.attr.get("aria-describedby")

  @describedby.setter
  def describedby(self, val):
    self.htmlObj.attr["aria-describedby"] = val

  @property
  def details(self):
    """
    Description:
    ------------
    Identifies the element that provides a detailed, extended description for the object. See related aria-describedby.

    The aria-details attribute references a single element that provides more detailed information than would normally be provided by aria-describedby. It enables assistive technologies to make users aware of the availability of an extended description as well as navigate to it. Authors SHOULD ensure the element referenced by aria-details is visible to all users.

    Related Pages:

			https://www.w3.org/TR/wai-aria-1.1/#aria-details
    """
    return self.htmlObj.attr.get("aria-details")

  @details.setter
  def details(self, val):
    self.htmlObj.attr["aria-details"] = val

  @property
  def disabled(self):
    """
    Description:
    ------------
    Indicates that the element is perceivable but disabled, so it is not editable or otherwise operable. See related aria-hidden and aria-readonly.

    For example, irrelevant options in a radio group may be disabled.
    Disabled elements might not receive focus from the tab order. For some disabled elements, applications might choose not to support navigation to descendants.
    In addition to setting the aria-disabled attribute, authors SHOULD change the appearance (grayed out, etc.) to indicate that the item has been disabled.

    Related Pages:

			https://www.w3.org/TR/wai-aria-1.1/#aria-disabled
    """
    return self.htmlObj.attr.get("aria-disabled", False)

  @disabled.setter
  def disabled(self, bool):
    self.htmlObj.attr["aria-disabled"] = bool

  @property
  def errormessage(self):
    """
    Description:
    ------------
    Identifies the element that provides an error message for the object. See related aria-invalid and aria-describedby.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-errormessage
    """
    return self.htmlObj.attr.get("aria-errormessage")

  @errormessage.setter
  def errormessage(self, val):
    self.htmlObj.attr["aria-errormessage"] = val

  @property
  def expanded(self):
    """
    Description:
    ------------
    Indicates whether the element, or another grouping element it controls, is currently expanded or collapsed.

    For example, this indicates whether a portion of a tree is expanded or collapsed.
    In other instances, this may be applied to page sections to mark expandable and collapsible regions that are flexible for managing content density.
    Simplifying the user interface by collapsing sections may improve usability for all, including those with cognitive or developmental disabilities.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-expanded
    """
    return self.htmlObj.attr.get("aria-expanded")

  @expanded.setter
  def expanded(self, bool):
    self.htmlObj.attr["aria-expanded"] = bool

  @property
  def flowto(self):
    """
    Description:
    ------------
    Identifies the next element (or elements) in an alternate reading order of content which, at the user's discretion, allows assistive technology to override the general default of reading in document source order.

    When aria-flowto has a single IDREF, it allows assistive technologies to, at the user's request, forego normal document reading order and go to the targeted object.
    However, when aria-flowto is provided with multiple IDREFS, assistive technologies SHOULD present the referenced elements as path choices.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-flowto
    """
    return self.htmlObj.attr.get("aria-flowto")

  @flowto.setter
  def flowto(self, val):
    self.htmlObj.attr["aria-flowto"] = val

  @property
  def haspopup(self):
    """
    Description:
    ------------
    Indicates the availability and type of interactive popup element, such as menu or dialog, that can be triggered by an element.

    A popup element usually appears as a block of content that is on top of other content.
    Authors MUST ensure that the role of the element that serves as the container for the popup content is menu, listbox, tree, grid, or dialog, and that the value of aria-haspopup matches the role of the popup container.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-haspopup
    """
    return self.htmlObj.attr.get("aria-haspopup", False)

  @haspopup.setter
  def haspopup(self, bool):
    self.htmlObj.attr["aria-haspopup"] = bool

  @property
  def hidden(self):
    """
    Description:
    ------------
    Indicates whether the element is exposed to an accessibility API. See related aria-disabled.

    User agents determine an element's hidden status based on whether it is rendered, and the rendering is usually controlled by CSS.
    For example, an element whose display property is set to none is not rendered. An element is considered hidden if it, or any of its ancestors are not rendered or have their aria-hidden attribute value set to true.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-hidden
    """
    return self.htmlObj.attr.get("aria-hidden")

  @hidden.setter
  def hidden(self, bool):
    self.htmlObj.attr["aria-hidden"] = json.dumps(bool)

  @property
  def invalid(self):
    """
    Description:
    ------------
    Indicates the entered value does not conform to the format expected by the application. See related aria-errormessage.

    If the value is computed to be invalid or out-of-range, the application author SHOULD set this attribute to true.
    User agents SHOULD inform the user of the error.
    Application authors SHOULD provide suggestions for corrections if they are known.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-invalid
    """
    return self.htmlObj.attr.get("aria-invalid")

  @invalid.setter
  def invalid(self, bool):
    self.htmlObj.attr["aria-invalid"] = bool

  @property
  def keyshortcuts(self):
    """
    Description:
    ------------
    Indicates keyboard shortcuts that an author has implemented to activate or give focus to an element.

    The value of the aria-keyshortcuts attribute is a space-delimited list of keyboard shortcuts that can be pressed to activate a command or textbox widget.
    The keys defined in the shortcuts represent the physical keys pressed and not the actual characters generated.
    Each keyboard shortcut consists of one or more tokens delimited by the plus sign ("+") representing zero or more modifier keys and exactly one non-modifier key that must be pressed simultaneously to activate the given shortcut.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-keyshortcuts
    """
    return self.htmlObj.attr.get("aria-keyshortcuts")

  @keyshortcuts.setter
  def keyshortcuts(self, val):
    self.htmlObj.attr["aria-keyshortcuts"] = val

  @property
  def label(self):
    """
    Description:
    ------------
    Defines a string value that labels the current element. See related aria-labelledby.

    The purpose of aria-label is the same as that of aria-labelledby.
    It provides the user with a recognizable name of the object.
    The most common accessibility API mapping for a label is the accessible name property.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-label
    """
    return self.htmlObj.attr.get("aria-label")

  @label.setter
  def label(self, val):
    self.htmlObj.attr["aria-label"] = val

  @property
  def labelledby(self):
    """
    Description:
    ------------
    Identifies the element (or elements) that labels the current element. See related aria-describedby.

    The purpose of aria-labelledby is the same as that of aria-label.
    It provides the user with a recognizable name of the object.
    The most common accessibility API mapping for a label is the accessible name property.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-label
    """
    return self.htmlObj.attr.get("aria-labelledby")

  @labelledby.setter
  def labelledby(self, val):
    self.htmlObj.attr["aria-labelledby"] = val

  @property
  def level(self):
    """
    Description:
    ------------
    Defines the hierarchical level of an element within a structure.

    This can be applied inside trees to tree items, to headings inside a document, to nested grids, nested tablists and to other structural items that may appear inside a container or participate in an ownership hierarchy.
    The value for aria-level is an integer greater than or equal to 1.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-level
    """
    return self.htmlObj.attr.get("aria-level")

  @level.setter
  def level(self, val):
    self.htmlObj.attr["aria-level"] = val

  @property
  def live(self):
    """
    Description:
    ------------
    Indicates that an element will be updated, and describes the types of updates the user agents, assistive technologies, and user can expect from the live region.

    The values of this attribute are expressed in degrees of importance.
    When regions are specified as polite, assistive technologies will notify users of updates but generally do not interrupt the current task, and updates take low priority.
    When regions are specified as assertive, assistive technologies will immediately notify the user, and could potentially clear the speech queue of previous updates.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-level
    """
    return self.htmlObj.attr.get("aria-live")

  @live.setter
  def live(self, val):
    self.htmlObj.attr["aria-live"] = val

  @property
  def modal(self):
    """
    Description:
    ------------
    Indicates whether an element is modal when displayed.

    The aria-modal attribute is used to indicate that the presence of a "modal" element precludes usage of other content on the page. For example, when a modal dialog is displayed, it is expected that the user's interaction is limited to the contents of the dialog, until the modal dialog loses focus or is no longer displayed.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-modal
    """
    return self.htmlObj.attr.get("aria-modal")

  @modal.setter
  def modal(self, val):
    self.htmlObj.attr["aria-modal"] = val

  @property
  def multiline(self):
    """
    Description:
    ------------
    Indicates whether a text box accepts multiple lines of input or only a single line.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-multiline
    """
    return self.htmlObj.attr.get("aria-multiline", False)

  @multiline.setter
  def multiline(self, bool):
    self.htmlObj.attr["aria-multiline"] = bool

  @property
  def multiselectable(self):
    """
    Description:
    ------------
    Indicates that the user may select more than one item from the current selectable descendants.

    Authors SHOULD ensure that selected descendants have the aria-selected attribute set to true, and selectable descendant have the aria-selected attribute set to false. Authors SHOULD NOT use the aria-selected attribute on descendants that are not selectable


    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-multiselectable
    """
    return self.htmlObj.attr.get("aria-multiselectable", False)

  @multiselectable.setter
  def multiselectable(self, val):
    self.htmlObj.attr["aria-multiselectable"] = val

  @property
  def orientation(self):
    """
    Description:
    ------------
    Indicates whether the element's orientation is horizontal, vertical, or unknown/ambiguous.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-multiselectable
    """
    return self.htmlObj.attr.get("aria-orientation")

  @orientation.setter
  def orientation(self, val):
    self.htmlObj.attr["aria-orientation"] = val

  @property
  def owns(self):
    """
    Description:
    ------------
    Identifies an element (or elements) in order to define a visual, functional, or contextual parent/child relationship between DOM elements where the DOM hierarchy cannot be used to represent the relationship. See related aria-controls.

    The value of the aria-owns attribute is a space-separated list of IDREFS that reference one or more elements in the document by ID. The reason for adding aria-owns is to expose a parent/child contextual relationship to assistive technologies that is otherwise impossible to infer from the DOM.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-owns
    """
    return self.htmlObj.attr.get("aria-owns")

  @owns.setter
  def owns(self, val):
    self.htmlObj.attr["aria-owns"] = val

  @property
  def placeholder(self):
    """
    Description:
    ------------
    Defines a short hint (a word or short phrase) intended to aid the user with data entry when the control has no value. A hint could be a sample value or a brief description of the expected format.

    Authors SHOULD NOT use aria-placeholder instead of a label as their purposes are different: The label indicates what kind of information is expected. The placeholder text is a hint about the expected value. See related aria-labelledby and aria-label.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-placeholder
    """
    return self.htmlObj.attr.get("aria-placeholder")

  @placeholder.setter
  def placeholder(self, val):
    self.htmlObj.attr["aria-placeholder"] = val

  @property
  def posinset(self):
    """
    Description:
    ------------
    Defines an element's number or position in the current set of listitems or treeitems. Not required if all elements in the set are present in the DOM. See related aria-setsize.

    If all items in a set are present in the document structure, it is not necessary to set this attribute, as the user agent can automatically calculate the set size and position for each item. However, if only a portion of the set is present in the document structure at a given moment, this property is needed to provide an explicit indication of an element's position

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-posinset
    """
    return self.htmlObj.attr.get("aria-posinset", 0)

  @posinset.setter
  def posinset(self, num):
    self.htmlObj.attr["aria-posinset"] = num

  @property
  def pressed(self):
    """
    Description:
    ------------
    Indicates the current "pressed" state of toggle buttons. See related aria-checked and aria-selected.

    Toggle buttons require a full press-and-release cycle to change their value. Activating it once changes the value to true, and activating it another time changes the value back to false. A value of mixed means that the values of more than one item controlled by the button do not all share the same value. Examples of mixed-state buttons are described in WAI-ARIA Authoring Practices [wai-aria-practices-1.1]. If the attribute is not present, the button is not a toggle button.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-pressed
    """
    return self.htmlObj.attr.get("aria-pressed", False)

  @pressed.setter
  def pressed(self, bool):
    self.htmlObj.attr["aria-pressed"] = bool

  @property
  def readonly(self):
    """
    Description:
    ------------
    Indicates that the element is not editable, but is otherwise operable. See related aria-disabled.

    This means the user can read but not set the value of the widget.
    Readonly elements are relevant to the user, and application authors SHOULD NOT restrict navigation to the element or its focusable descendants. Other actions such as copying the value of the element are also supported. This is in contrast to disabled elements, to which applications might not allow user navigation to descendants.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-readonly
    """
    return self.htmlObj.attr.get("aria-readonly", False)

  @readonly.setter
  def readonly(self, bool):
    self.htmlObj.attr["aria-readonly"] = bool

  @property
  def relevant(self):
    """
    Description:
    ------------
    Indicates what notifications the user agent will trigger when the accessibility tree within a live region is modified. See related aria-atomic.

    The attribute is represented as a space delimited list of the following values: additions, removals, text; or a single catch-all value all.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-relevant
    """
    return self.htmlObj.attr.get("aria-relevant", False)

  @relevant.setter
  def relevant(self, bool):
    self.htmlObj.attr["aria-relevant"] = bool

  @property
  def relevant(self):
    """
    Description:
    ------------
    Indicates that user input is required on the element before a form may be submitted.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-relevant
    """
    return self.htmlObj.attr.get("aria-relevant", False)

  @relevant.setter
  def relevant(self, bool):
    self.htmlObj.attr["aria-relevant"] = bool

  @property
  def roledescription(self):
    """
    Description:
    ------------
    Defines a human-readable, author-localized description for the role of an element.

    Some assistive technologies, such as screen readers, present the role of an element as part of the user experience. Such assistive technologies typically localize the name of the role, and they may customize it as well. Users of these assistive technologies depend on the presentation of the role name, such as "region," "button," or "slider," for an understanding of the purpose of the element and, if it is a widget, how to interact with it.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-roledescription
    """
    return self.htmlObj.attr.get("aria-roledescription")

  @roledescription.setter
  def roledescription(self, val):
    self.htmlObj.attr["aria-roledescription"] = val

  @property
  def rowindex(self):
    """
    Description:
    ------------
    Defines an element's row index or position with respect to the total number of rows within a table, grid, or treegrid. See related aria-rowcount and aria-rowspan.

    If all of the rows are present in the DOM, it is not necessary to set this attribute as the user agent can automatically calculate the index of each row. However, if only a portion of the rows is present in the DOM at a given moment, this attribute is needed to provide an explicit indication of each row's position with respect to the full table.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-rowindex
    """
    return self.htmlObj.attr.get("aria-rowindex")

  @rowindex.setter
  def rowindex(self, num):
    self.htmlObj.attr["aria-roledescription"] = num

  @property
  def rowspan(self):
    """
    Description:
    ------------
    Defines the number of rows spanned by a cell or gridcell within a table, grid, or treegrid. See related aria-rowindex and aria-colspan.

    This attribute is intended for cells and gridcells which are not contained in a native table. When defining the row span of cells or gridcells in a native table, authors SHOULD use the host language's attribute instead of aria-rowspan. If aria-rowspan is used on an element for which the host language provides an equivalent attribute, user agents MUST ignore the value of aria-rowspan and instead expose the value of the host language's attribute to assistive technologies.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-rowspan
    """
    return self.htmlObj.attr.get("aria-rowspan")

  @rowspan.setter
  def rowspan(self, num):
    self.htmlObj.attr["aria-rowspan"] = num

  @property
  def selected(self):
    """
    Description:
    ------------
    Indicates the current "selected" state of various widgets. See related aria-checked and aria-pressed.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-selected
    """
    return self.htmlObj.attr.get("aria-selected", False)

  @selected.setter
  def selected(self, bool):
    self.htmlObj.attr["aria-selected"] = bool

  @property
  def setsize(self):
    """
    Description:
    ------------
    Defines the number of items in the current set of listitems or treeitems. Not required if all elements in the set are present in the DOM. See related aria-posinset.

    This property is marked on the members of a set, not the container element that collects the members of the set.
    To orient the user by saying an element is "item X out of Y," the assistive technologies would use X equal to the aria-posinset attribute and Y equal to the aria-setsize attribute.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-setsize
    """
    return self.htmlObj.attr.get("aria-setsize")

  @setsize.setter
  def setsize(self, num):
    self.htmlObj.attr["aria-setsize"] = num

  @property
  def sort(self):
    """
    Description:
    ------------
    Defines the maximum allowed value for a range widget.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-sort
    """
    return self.htmlObj.attr.get("aria-sort", False)

  @sort.setter
  def sort(self, bool):
    self.htmlObj.attr["aria-sort"] = bool

  @property
  def valuemin(self):
    """
    Description:
    ------------
    Defines the minimum allowed value for a range widget.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-valuemin
    """
    return self.htmlObj.attr.get("aria-valuemin")

  @valuemin.setter
  def valuemin(self, num):
    self.htmlObj.attr["aria-valuemin"] = num

  @property
  def valuemax(self):
    """
    Description:
    ------------
    Defines the maximum allowed value for a range widget.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-valuemax
    """
    return self.htmlObj.attr.get("aria-valuemax")

  @valuemax.setter
  def valuemax(self, num):
    self.htmlObj.attr["aria-valuemax"] = num

  @property
  def valuenow(self):
    """
    Description:
    ------------
    Defines the current value for a range widget. See related aria-valuetext.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-valuenow
    """
    return self.htmlObj.attr.get("aria-valuenow")

  @valuenow.setter
  def valuenow(self, num):
    self.htmlObj.attr["aria-valuenow"] = num

  @property
  def valuetext(self):
    """
    Description:
    ------------
    Defines the human readable text alternative of aria-valuenow for a range widget.

    Related Pages:

      https://www.w3.org/TR/wai-aria-1.1/#aria-valuetext
    """
    return self.htmlObj.attr.get("aria-valuetext")

  @valuetext.setter
  def valuetext(self, text):
    self.htmlObj.attr["aria-valuetext"] = text





class D3(object):

  def __init__(self, context):
    self.parent = context

  def table(self, records=None, header=None, width=(100, '%'), height=(None, 'px'), htmlCode=None, options=None,
            profile=None):
    """

    :param records:
    :param header:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param profile:
    """
    table = self.parent.context.rptObj.ui.div(width=width, height=height, htmlCode=htmlCode, options=options,
                                              profile=profile)
    d3_table = table.js.d3.select("#%s" % table.htmlCode, varName='d3Table').rappend('table')
    if header is None and records is not None:
      header = list(records[0].keys())
    table.onReady(
      [d3_table.toStr(),
       d3_table.rappend('thead').rappend('tr')
         .selectAll('th').data(header).enter().rappend('th').text("ƒ('head')").toStr(),
       d3_table.rappend('tbody').selectAll('tr').data(records).enter()
         .rappend('tr').selectAll('td').dataRecord(header).enter().rappend('td').html().toStr()
       ]
    )
    return table




from epyk.core.html import graph


class Vis2D(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "Vis"

  def line(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None,
           htmlCode=None):
    """
    Description:
    -----------
    Graph2d is an interactive visualization chart to draw data in a 2D graph. You can freely move and zoom in the graph by dragging and scrolling in the window.

    Graph2d uses HTML DOM and SVG for rendering. This allows for flexible customization using css styling.

    elated Pages:
    --------------
    http://www.chartjs.org/
    https://visjs.github.io/vis-timeline/examples/graph2d/16_bothAxisTitles.html

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param height:
    :param htmlCode:
    """
    data = self.parent.context.rptObj.data.vis.xy(record, y_columns, x_axis)
    line_chart = graph.GraphVis.ChartLine(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    #line_chart.groups.add(sorted(list(labels)))
    for d in data:
      line_chart.add_items(d)
    return line_chart

  def scatter(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"),
              options=None, htmlCode=None):
    """
    Description:
    -----------

    elated Pages:
    --------------
    http://www.chartjs.org/

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param height:
    :param htmlCode:
    """
    data = self.parent.context.rptObj.data.vis.xy(record, y_columns, x_axis)
    line_chart = graph.GraphVis.ChartScatter(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.options.height = height[0]
    #line_chart.groups.add(sorted(list(labels)))
    for d in data:
      line_chart.add_items(d)
    return line_chart

  def bar(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"),
          options=None, htmlCode=None):
    """
    Description:
    -----------

    elated Pages:
    --------------
    http://www.chartjs.org/

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param height:
    :param htmlCode:
    """
    data = self.parent.context.rptObj.data.vis.xy(record, y_columns, x_axis)
    line_chart = graph.GraphVis.ChartBar(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.options.height = height[0]
    for d in data:
      line_chart.add_items(d)
    return line_chart

  def timeline(self, record=None, start=None, content=None, end=None, type=None, group=None, profile=None,
               width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    -----------

    Related Pages:
http://www.chartjs.org/

    Attributes:
    ----------
    :param record:
    :param start:
    :param content:
    :param end:
    :param type:
    :param group:
    :param profile:
    :param height:
    :param options:
    :param htmlCode:
    """

    data = self.parent.context.rptObj.data.vis.timeline(record, start, content, end, type, group, options)
    line_chart = graph.GraphVis.ChartTimeline(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.options.height = height[0]
    line_chart.options.editable = True
    line_chart.options.showCurrentTime = True
    if data['datasets']:
      line_chart.add_items(data['datasets'])
    return line_chart

  def network(self, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    line_chart = graph.GraphVis.ChartNetwork(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.options.height = height[0]
    return line_chart


class Vis3D(object):

  def __init__(self, context):
    self.parent = context
    self.chartFamily = "Vis"

  def line(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, width=(100, "%"), height=(400, "px"), options=None, htmlCode=None):
    """
    Description:
    -----------

    :param record:
    :param y_columns:
    :param x_axis:
    :param z_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    data = self.parent.context.rptObj.data.vis.xyz(record, y_columns, x_axis, z_axis)
    line_chart = graph.GraphVis.Chart3DLine(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    for d in data:
      line_chart.add_items(d)
    return line_chart

  def bar(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, width=(100, "%"), height=(400, "px"), options=None, htmlCode=None):
    """
    Description:
    -----------

    :param record:
    :param y_columns:
    :param x_axis:
    :param z_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    data = self.parent.context.rptObj.data.vis.xyz(record, y_columns, x_axis, z_axis)
    line_chart = graph.GraphVis.Chart3DBar(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    for d in data:
      line_chart.add_items(d)
    return line_chart

  def surface(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, width=(100, "%"), height=(400, "px"), options=None, htmlCode=None):
    """
    Description:
    -----------

    :param record:
    :param y_columns:
    :param x_axis:
    :param z_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    data = self.parent.context.rptObj.data.vis.xyz(record, y_columns, x_axis, z_axis)
    line_chart = graph.GraphVis.Chart3D(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    for d in data:
      line_chart.add_items(d)
    return line_chart

  def scatter(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, width=(100, "%"), height=(400, "px"), options=None, htmlCode=None):
    """
    Description:
    -----------

    :param record:
    :param y_columns:
    :param x_axis:
    :param z_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    data = self.parent.context.rptObj.data.vis.xyz(record, y_columns, x_axis, z_axis)
    line_chart = graph.GraphVis.Chart3DScatter(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    for d in data:
      line_chart.add_items(d)
    return line_chart


class Vis(Vis2D):

  def __init__(self, context):
    super(Vis, self).__init__(context)
    self._3d = Vis3D(context)

  def surface(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, width=(100, "%"), height=(400, "px"), options=None, htmlCode=None):
    return self._3d.surface(record, y_columns, x_axis, z_axis, profile, width, height, options, htmlCode)

  def bar3d(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, width=(100, "%"), height=(400, "px"), options=None, htmlCode=None):
    return self._3d.bar(record, y_columns, x_axis, z_axis, profile, width, height, options, htmlCode)

  def scatter3d(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, width=(100, "%"), height=(400, "px"), options=None, htmlCode=None):
    return self._3d.scatter(record, y_columns, x_axis, z_axis, profile, width, height, options, htmlCode)

  def line3d(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, width=(100, "%"), height=(400, "px"), options=None, htmlCode=None):
    return self._3d.line(record, y_columns, x_axis, z_axis, profile, width, height, options, htmlCode)



"""
CSS Style module for the List components
"""

from epyk.core.css.styles.classes import CssStyle


class CssBasicList(CssStyle.Style):
  _attrs = {'padding': '15px', 'font-size': '16px', 'font-weight': 'bold'}
  _selectors = {'child': 'ul li:first-child'}

  def customize(self):
    self.css({"background": self.rptObj.theme.colors[7], "color": self.rptObj.theme.greys[0]})


class CssBasicListItems(CssStyle.Style):
  _attrs = {'padding': '15px 0 0 0', 'display': 'block'}

  def customize(self):
    self.css({"color": self.rptObj.theme.greys[-1], "border-bottom": "1px solid %s" % self.rptObj.theme.colors[7]})
    self.hover.css({'color': self.rptObj.theme.colors[8]})


class CssBasicListItemsSelected(CssStyle.Style):

  def customize(self):
    self.css({"color": self.rptObj.theme.success[1]})


class CssBasicListItemsDisabled(CssStyle.Style):
  _attrs = {'padding': '15px 0 0 0', 'display': 'block', 'cursor': 'not-allowed'}

  def customize(self):
    self.css({"background-color": self.rptObj.theme.greys[7], "color": self.rptObj.theme.greys[4]})


class CssSquareList(CssStyle.Style):
  _attrs = {'list-style': 'none', 'text-align': 'justify'}
  _before = {'content': "'\\f0c8'", 'font-family': "'Font Awesome 5 Free'", 'padding': '0 5px 0 0'}
  _selectors = {'child': 'li'}

  def customize(self):
    self.before.cess({"color": self.rptObj.theme.colors[7]})


class CssListBase(CssStyle.Style):
  _attrs = {'width': '142px', 'min-height': '20px', 'list-style-type': 'none', 'margin': 0, 'padding': '5px 0 0 0',
            'float': 'left', 'margin-right': '10px'}

  def customize(self):
    self.hover.css({"border": "1px solid %s" % self.rptObj.theme.greys[9]})


class CssListLiBase(CssStyle.Style):
  _attrs = {'margin': '0 5px 5px 5px', 'padding': '5px', 'width': '120px'}


class CssListNoDecoration(CssStyle.Style):
  _attrs = {'list-style': 'none', 'padding': '0 0 0 5px'}


class CssListLiUlContainer(CssStyle.Style):
  _attrs = {'display': 'none', 'width': '100%', 'padding': 0, 'margin': 0, 'overflow': 'hidden',
            'transition': 'height .2s ease-in-out'}
  cssId = {'direct': 'ul'}


class CssListLiSubItem(CssStyle.Style):
  _attrs = {'display': 'table', 'width': '100%', 'height': '30px', 'vertical-align': 'middle',
            'list-style-type': 'none', 'float': 'left', 'font-size': '12px', 'margin': 0}

  def customize(self):
    self.hover.css({"background": self.rptObj.theme.greys[2]})




from epyk.core import html


class Modals(object):
  def __init__(self, context):
    self.context = context

  def forms(self, html_objs, action, method, header=None, footer=None, helper=None):
    """
    Simple interface to create an html form within a modal

    Usage::

      d = rptObj.ui.fields.today('test')
      i = rptObj.ui.fields.input(placeholder='test2', label='test1')
      i2 = rptObj.ui.fields.input('test3', label='test2')
      form_modal = rptObj.ui.modals.forms([d, i, i2], "http://127.0.0.1:5000", "POST")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Modal`
      - :class:`epyk.core.html.HtmlContainer.Form`

    Related Pages:

        https://www.w3schools.com/w3css/w3css_modal.asp

    Attributes:
    ----------
    :param html_objs list:
    :param method:
    :param helper:
    :return:
    """
    if not type(html_objs) == list:
      html_objs = [html_objs]
    form = html.HtmlContainer.Form(self.context.rptObj, html_objs, helper)
    form.submit(method, action)
    modal = html.HtmlContainer.Modal(self.context.rptObj, [], header, footer, False, helper)
    modal += form
    modal.form = form
    return modal

  def disclaimer(self, disc_list, header=None, footer=None, submit=True, validation_text='AGREE', action=None, add_buttons=None, to_html=True, helper=None):
    """
    Disclaimer that will appear as a modal

    Usage::

      privacy_title = rptObj.ui.texts.title('A privacy reminder', 2)
      p1 = rptObj.ui.texts.paragraph('''Scroll down and click “%s” when you’re ready to continue, or explore other options on this page.''' % rptObj.ui.tags.strong('''I agree''', options={'managed': False}))
      disc = rptObj.ui.modals.disclaimer([privacy_title, p1])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Modal`
      - :class:`epyk.core.html.HtmlContainer.Row`
      - :class:`epyk.core.html.HtmlButton.Button`

    :param disc_list:
    :param header:
    :param footer:
    :param submit:
    :param validation_text:
    :param action:
    :param add_buttons:
    :param to_html:
    :param helper:
    :return:
    """
    for obj in disc_list:
      obj.css({'margin': '40px', 'width': 'auto', 'text-align': 'justify'})

    modal = html.HtmlContainer.Modal(self.context.rptObj, disc_list, header, footer, False, helper)
    modal.col.css({'width': '450px', 'height':'700px'})
    if add_buttons or submit:
      submitRow = self.context.rptObj.ui.row([]) if not add_buttons else self.context.rptObj.ui.row(add_buttons)
      if submit:
        submitBtn = self.context.rptObj.ui.buttons.important(validation_text)
        if action:
          submitBtn.click(action)
        else:
          submitBtn.click(modal.close())
        submitRow + submitBtn
      modal.col + submitRow
    return modal

  def dialog(self, text, width=(100, '%'), height=(20, 'px'), htmlCode=None, attrs=None,
                  helper=None, options=None, profile=None):
    """
    Description:
    ------------
    Simple Jquery UI modal with a text

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlEvent.Dialog`

    Related Pages:

        https://jqueryui.com/dialog/

    Attributes:
    ----------
    :param text:
    :param width:
    :param height:
    :param htmlCode:
    :param attrs:
    :param helper:
    :param options:
    :param profile:
    """
    html_pr = html.HtmlEvent.Dialog(self.context.rptObj, text, width, height,  attrs or {}, helper,
                                         options or {}, htmlCode, profile)
    return html_pr




from epyk.core.css.styles import GrpCls
from epyk.core.css import Classes


class ClassNetworkBot(GrpCls.ClassHtml):

  def __init__(self, htmlObj):
    super(ClassNetworkBot, self).__init__(htmlObj)
    self._css_bubble, self._css_bubble_content, self._css_bubble_arrow = None, None, None
    self.classList['other'].add(self.cls_bubble)
    self.classList['other'].add(self.cls_bubble_content)
    self.classList['other'].add(self.cls_bubble_arrow)

  @property
  def cls_bubble(self):
    """

    :rtype: Classes.CatalogInput.CatalogDate
    """
    if self._css_bubble is None:
      self._css_bubble = Classes.CatalogDiv.CatalogDiv(self.htmlObj._report, self.classList['other']).bubble_container()
    return self._css_bubble

  @property
  def cls_bubble_content(self):
    """

    :rtype: Classes.CatalogInput.CatalogDate
    """
    if self._css_bubble_content is None:
      self._css_bubble_content = Classes.CatalogDiv.CatalogDiv(self.htmlObj._report, self.classList['other']).bubble_content()
    return self._css_bubble_content

  @property
  def cls_bubble_arrow(self):
    """

    :rtype: Classes.CatalogInput.CatalogDate
    """
    if self._css_bubble_arrow is None:
      self._css_bubble_arrow = Classes.CatalogDiv.CatalogDiv(self.htmlObj._report, self.classList['other']).bubble_arrow()
    return self._css_bubble_arrow




import datetime
import json

from epyk.core.html import Html
from epyk.core.html import Defaults
from epyk.core.html.options import OptInputs

#
from epyk.core.js import JsUtils
from epyk.core.js.objects import JsComponents
from epyk.core.js.packages import JsQuery
from epyk.core.js.packages import JsTimepicker
from epyk.core.js.packages import JsQueryUi
from epyk.core.js.html import JsHtmlField
from epyk.core.js.html import JsHtmlJqueryUI

# The list of CSS classes
from epyk.core.css.styles import GrpClsInput


class Output(Html.Html):
  name = 'Output'

  def __str__(self):
    return '<output %(strAttr)s>%(val)s</output>' % {'strAttr': self.get_attrs(pyClassNames=self.style.get_classes()), 'val': self.val}


class Input(Html.Html):
  name = 'Input'

  def __init__(self, report, text, placeholder, width, height, htmlCode, options, attrs, profile):
    super(Input, self).__init__(report, text, htmlCode=htmlCode, css_attrs={"width": width, "height": height, 'box-sizing': 'border-box'},
                                profile=profile, options=options)
    value = text['value'] if isinstance(text, dict) else text
    self.set_attrs(attrs={"placeholder": placeholder, "type": "text", "value": value, "spellcheck": False})
    self.set_attrs(attrs=attrs)
    self.__options, self.__focus = OptInputs.OptionsInput(self, options), False

  @property
  def options(self):
    """
    Description:
    -----------
    Property to set all the input component properties

    :rtype: OptInputs.OptionsInput
    """
    return self.__options

  @property
  def style(self):
    """
    Description:
    ------------
    Property to the CSS Style of the component

    :rtype: GrpClsInput.ClassInput
    """
    if self._styleObj is None:
      self._styleObj = GrpClsInput.ClassInput(self)
    return self._styleObj

  @property
  def _js__builder__(self):
    return '''
      if(typeof options.css !== 'undefined'){for(var k in options.css){htmlObj.style[k] = options.css[k]}}
      if(typeof options.formatMoney !== 'undefined'){ htmlObj.value = accounting.formatMoney(data, 
        options.formatMoney.symbol, options.formatMoney.digit, 
        options.formatMoney.thousand, options.formatMoney.decimal)}
      else if(typeof options.formatNumber !== 'undefined'){ htmlObj.value = accounting.formatNumber(data,
        options.formatNumber.digit, options.formatNumber.thousand)}
      else if(typeof options.toFixed !== 'undefined'){ htmlObj.value = accounting.toFixed(data, options.toFixed) }
      else { htmlObj.value = data; }
      '''

  def focus(self, jsFncs=None, profile=False, options=None):
    """
    Description:
    -----------
    Action on focus

    Attributes:
    ----------
    :param jsFncs: List or String with the Javascript events
    :param profile: Boolean to add the Javascript fragment to profile
    :param options: Python dictionary with special options (shortcuts) for the component
    """
    self.__focus = True
    if jsFncs is None:
      jsFncs = []
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    if options is not None:
      if options.get("reset", False):
        jsFncs.append(self.dom.empty())
      if options.get("select", False):
        jsFncs.append(self.dom.select())
    if self.options.reset:
      jsFncs.append(self.dom.empty())
    if self.options.select:
      jsFncs.append(self.dom.select())
    return self.on("focus", jsFncs, profile)

  def validation(self, pattern, required=True):
    """
    Description:
    -----------
    Add validation rules on the input component

    Usage::

      input.validation(pattern="[0-9]{5}")

    Attributes:
    ----------
    :param pattern: String.
    :return: Self to allow the chaining
    """
    self.attr["pattern"] = pattern
    if required:
      self.attr["required"] = None
    self.style.add_classes.input.is_valid()
    return self

  def enter(self, jsFncs, profile=False, source_event=None):
    """
    Description:
    ------------
    Add an javascript action when the key enter is pressed on the keyboard

    Usage::

      htmlObj.input(placeholder="Put your tag").enter("alert()")

    Attributes:
    ----------
    :param jsFncs:
    :param profile:
    :param source_event:

    :return: The python object itself
    """

    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    jsFncs.append(self.dom.select())
    self.keydown.enter(jsFncs, profile, source_event=source_event)
    return self

  def readonly(self, flag=True):
    """

    :param flag:
    :return:
    """
    if flag:
      self.attr["readonly"] = "readonly"
    else:
      if "readonly" in self.attr:
        del self.attr["readonly"]
    return self

  def __str__(self):
    if not self.__focus and (self.options.reset or self.options.select):
      self.focus()
    if 'css' in self._jsStyles:
      self.css(self._jsStyles['css'])
    return '<input %(strAttr)s />' % {'strAttr': self.get_attrs(pyClassNames=self.style.get_classes())}


class InputRadio(Input):
  name = 'Input'

  def __init__(self, report, flag, group_name, placeholder, width, height, htmlCode, options, attrs, profile):
    super(InputRadio, self).__init__(report, "", placeholder, width, height, htmlCode, options, attrs, profile)
    self.set_attrs({"type": 'radio'})
    if flag:
      self.set_attrs({"checked": json.dumps(flag)})
    if group_name is not None:
      self.set_attrs({"name": group_name})


class AutoComplete(Input):
  name = 'Input Time'
  requirements = ('jqueryui', )

  def __init__(self, report, text, placeholder, width, height, htmlCode, options, attrs, profile):
    if text is None:
      text = str(datetime.datetime.now()).split(" ")[1].split(".")[0]
    super(AutoComplete, self).__init__(report, text, placeholder, width, height, htmlCode, options, attrs, profile)
    self.__options = OptInputs.OptionAutoComplete(self, options)

  @property
  def _js__builder__(self):
    return "jQuery(htmlObj).autocomplete(options)"

  @property
  def options(self):
    """
    Description:
    ------------
    Property to set all the input timepicker component properties

    Related Pages:

			https://timepicker.co/options/

    :rtype: OptInputs.OptionAutoComplete
    """
    return self.__options

  @property
  def js(self):
    """
    Description:
    -----------

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsQueryUi.Autocomplete
    """
    if self._js is None:
      self._js = JsQueryUi.Autocomplete(self, report=self._report)
    return self._js

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<input %(strAttr)s />' % {'strAttr': self.get_attrs(pyClassNames=self.style.get_classes())}


class InputTime(Input):
  name = 'Input Time'
  requirements = ('timepicker', )

  def __init__(self, report, text, placeholder, width, height, htmlCode, options, attrs, profile):
    if text is None:
      text = str(datetime.datetime.now()).split(" ")[1].split(".")[0]
    super(InputTime, self).__init__(report, text, placeholder, width, height, htmlCode, options, attrs, profile)
    self.__options = OptInputs.OptionsTimePicker(self, options)

  @property
  def options(self):
    """
    Description:
    ------------
    Property to set all the input timepicker component properties

    Related Pages:

			https://timepicker.co/options/

    :rtype: OptInputs.OptionsTimePicker
    """
    return self.__options

  @property
  def style(self):
    """
    Description:
    ------------
    Property to the CSS Style of the component

    :rtype: GrpClsInput.ClassInputTime
    """
    if self._styleObj is None:
      self._styleObj = GrpClsInput.ClassInputTime(self)
    return self._styleObj

  @property
  def dom(self):
    """
    Description:
    ------------
    The Javascript Dom object

    :rtype: JsHtmlJqueryUI.JsHtmlTimePicker
    """
    if self._dom is None:
      self._dom = JsHtmlJqueryUI.JsHtmlTimePicker(self, report=self._report)
    return self._dom

  @property
  def js(self):
    """
    Description:
    -----------

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsTimepicker.Timepicker
    """
    if self._js is None:
      self._js = JsTimepicker.Timepicker(self, report=self._report)
    return self._js

  @property
  def _js__builder__(self):
    """
    else {
        if (data.value == ''){data.time = new Date()} else{data.time = data.value};
        if (data.options._change.length > 0) {data.options.change = function(time){
            let data = {event_val: time.getHours() +':'+ time.getMinutes() +':'+ time.getSeconds(), event_code: htmlCode};
            eval(data.options._change.join(";"))}};
        jQuery(htmlObj).timepicker(data.options); jQuery(htmlObj).timepicker('setTime', data.time)}

    :return:
    """
    return '''
      if (typeof data == "string"){%(jqId)s.timepicker('setTime', data)} 
      %(jqId)s.timepicker(options); ''' % {"jqId": JsQuery.decorate_var("htmlObj", convert_var=False)}

  def change(self, jsFnc):
    """
    Description:
    -----------
    Event triggerd when the value of the input field changes. A Date object containing the selected time is passed as the first argument of the callback.
    Note: the variable time is a function parameter received in the Javascript side

    Related Pages:

			https://timepicker.co/options/

    :param jsFnc:
    """
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    self._jsStyles["change"] = "function(time){ %s }" % JsUtils.jsConvertFncs(jsFnc, toStr=True)
    return self

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<input %(strAttr)s />' % {'strAttr': self.get_attrs(pyClassNames=self.style.get_classes())}


class InputDate(Input):
  requirements = ('jqueryui', )
  name = 'Input Time'

  def __init__(self, report, records, placeholder, width, height, htmlCode, options, attrs, profile):
    super(InputDate, self).__init__(report, records, placeholder, width, height, htmlCode, options, attrs, profile)
    self.__options = OptInputs.OptionsDatePicker(self, options)

  @property
  def options(self):
    """
    Description:
    ------------
    Property to set all the input Datepicker component properties

    Related Pages:

			https://timepicker.co/options/

    :rtype: OptInputs.OptionsDatePicker
    """
    return self.__options

  @property
  def js(self):
    """
    Description:
    -----------

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsQueryUi.Datepicker
    """
    if self._js is None:
      self._js = JsQueryUi.Datepicker(self, report=self._report)
    return self._js

  @property
  def style(self):
    """
    Description:
    ------------
    Property to the CSS Style of the component

    :rtype: GrpClsInput.ClassInput
    """
    if self._styleObj is None:
      self._styleObj = GrpClsInput.ClassInputDate(self)
    return self._styleObj

  @property
  def dom(self):
    """
    Description:
    ------------
    The Javascript Dom object

    :rtype: JsHtmlJqueryUI.JsHtmlDatePicker
    """
    if self._dom is None:
      self._dom = JsHtmlJqueryUI.JsHtmlDatePicker(self, report=self._report)
    return self._dom

  def excluded_dates(self, dts=None, jsFncs=None):
    """

    Attributes:
    ----------
    :param dts:
    :param jsFncs:
    """
    self._jsStyles['beforeShowDay'] = '''function (date) {
            var utc = date.getTime() - date.getTimezoneOffset()*60000; var newDate = new Date(utc); const dts = %(dts)s;
            %(jsFnc)s; if(dts.includes(newDate.toISOString().split('T')[0])){return [false, '', '']} else {return [true, '', '']}
          }''' % {"dts": json.dumps(dts or []), 'jsFnc': JsUtils.jsConvertFncs(jsFncs, toStr=True)}

  def included_dates(self, dts=None, jsFncs=None):
    """

    Attributes:
    ----------
    :param dts:
    :param jsFncs:
    """
    self._jsStyles['beforeShowDay'] = '''function (date) {
        var utc = date.getTime() - date.getTimezoneOffset()*60000; var newDate = new Date(utc); const dts = %(dts)s;
        %(jsFnc)s; if(!dts.includes(newDate.toISOString().split('T')[0])){return [false, '', '']} else {return [true, '', '']}
      }''' % {"dts": json.dumps(dts or []), 'jsFnc': JsUtils.jsConvertFncs(jsFncs, toStr=True)}

  @property
  def _js__builder__(self):
    '''
    if ((typeof data.options.selectedDts !== "undefined") && (data.options.selectedDts.length > 0)){
        var selectedDt = {};
        data.options.selectedDts.forEach(function(dt){var jsDt = new Date(dt); selectedDt[jsDt.toISOString().split('T')[0]] = jsDt}) ;
        if (data.options.excludeDts === true){
          function renderCalendarCallbackExc(intDate) {var utc = intDate.getTime() - intDate.getTimezoneOffset()*60000; var newDate = new Date(utc); var Highlight = selectedDt[newDate.toISOString().split('T')[0]]; if(Highlight){return [false, '', '']} else {return [true, '', '']}};
          data.options.beforeShowDay = renderCalendarCallbackExc;
        } else{
          function renderCalendarCallback(intDate) {var utc = intDate.getTime() - intDate.getTimezoneOffset()*60000; var newDate = new Date(utc); var Highlight = selectedDt[newDate.toISOString().split('T')[0]]; if(Highlight){return [true, "%s", '']} else {return [false, '', '']}};
          data.options.beforeShowDay = renderCalendarCallback;};
        delete data.options.selectedDts};
      jQuery(htmlObj).datepicker(data.options).datepicker('setDate', data.value)
    :return:
    '''
    return '''%(jqId)s.datepicker(options).datepicker('setDate', data)''' % {"jqId": JsQuery.decorate_var("htmlObj", convert_var=False)}

  def __str__(self):
    # Javascript builder is mandatory for this object
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    if self.options.inline:
      return '<div %(strAttr)s></div>' % {'strAttr': self.get_attrs(pyClassNames=self.style.get_classes())}

    return '<input %(strAttr)s />' % {'strAttr': self.get_attrs(pyClassNames=self.style.get_classes())}


class InputInteger(Input):
  name = 'Input Number'

  def __init__(self, report, text, placeholder, width, height, htmlCode, options, attrs, profile):
    super(InputInteger, self).__init__(report, text, placeholder, width, height, htmlCode, options, attrs, profile)
    self.__options = OptInputs.OptionsInputInteger(self, options)

  @property
  def options(self):
    """
    Description:
    -----------
    Property to set all the input component properties

    :rtype: OptInputs.OptionsInputInteger
    """
    return self.__options

  def quantity(self):
    factors, strFcts = {'K': 1000, 'M': 1000000, 'B': 1000000000}, []
    for f, val in factors.items():
      strFcts.append("if(event.key.toUpperCase() == '%s'){this.value *= %s}" % (f, val))
    self.on('keydown', ";".join(strFcts))
    return self


class InputRange(Input):
  name = 'Input Range'

  def __init__(self, report, text, min, max, step, placeholder, width, height, htmlCode, options, attrs, profile):
    super(InputRange, self).__init__(report, text, placeholder, width, height, htmlCode, options, attrs, profile)
    self.__options = OptInputs.OptionsInputRange(self, options)
    #
    self.input = report.ui.inputs.input(text, width=(None, "px"), placeholder=placeholder).css({"vertical-align": 'middle'})
    self.append_child(self.input)
    #
    self.input.set_attrs(attrs={"type": "range", "min": min, "max": max, "step": step})
    if self.options.output:
      self.output = self._report.ui.inputs._output(text).css({
        "width": '15px', "text-align": 'center', "margin-left": '2px', 'color': self._report.theme.success[1]})
      self.append_child(self.output)
      self.input.set_attrs(attrs={"oninput": "%s.value=this.value" % self.output.htmlCode})
    self.css({"display": 'inline-block', "vertical-align": 'middle', "line-height": '%spx' % Defaults.LINE_HEIGHT})

  @property
  def options(self):
    """
    Property to set input range properties

    :rtype: OptSelect.OptionsInputRange
    """
    return self.__options

  @property
  def style(self):
    """
    Description:
    ------------
    Property to the CSS Style of the component

    :rtype: GrpClsInput.ClassInputRange
    """
    if self._styleObj is None:
      self._styleObj = GrpClsInput.ClassInputRange(self)
    return self._styleObj#

  def __str__(self):
    if hasattr(self, 'output'):
      self.output.css({"display": 'inline-block'})
    return '<div %(strAttr)s></div>' % {'strAttr': self.get_attrs(pyClassNames=self.style.get_classes())}


class Field(Html.Html):
  name = 'Field'

  def __init__(self, report, input, label, placeholder, icon, width, height, htmlCode, helper, profile):
    super(Field, self).__init__(report, "", htmlCode=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    # Add the component predefined elements
    self.add_label(label)
    self.add_helper(helper, css={"line-height": '%spx' % Defaults.LINE_HEIGHT})
    # add the input item
    self.input = input
    self.append_child(self.input)
    self.add_icon(icon, position="after", css={"margin-left": '5px', 'color': self._report.theme.success[1]})
    self.css({"margin-top": '2px'})

  @property
  def dom(self):
    """
    Description:
    -----------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtmlField.JsHtmlFields
    """
    if self._dom is None:
      self._dom = JsHtmlField.JsHtmlFields(self, report=self._report)
    return self._dom

  def __str__(self):
    str_div = "".join([v.html() if hasattr(v, 'html') else v for v in self.val])
    return "<div %s>%s%s</div>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_div, self.helper)


class FieldInput(Field):
  name = 'Field Input'

  def __init__(self, report, value, label, placeholder, icon, width, height, htmlCode, helper, options, profile):
    input = report.ui.inputs.input(value, width=(None, "%"), placeholder=placeholder, options=options)
    super(FieldInput, self).__init__(report, input, label, placeholder, icon, width, height, htmlCode, helper, profile)


class FieldAutocomplete(Field):
  name = 'Field Autocomplete'

  def __init__(self, report, value, label, placeholder, icon, width, height, htmlCode, helper, options, profile):
    input = report.ui.inputs.autocomplete(value, width=(None, "%"), placeholder=placeholder, options=options)
    super(FieldAutocomplete, self).__init__(report, input, label, placeholder, icon, width, height, htmlCode, helper, profile)

  def change(self, jsFnc, profile=None):
    """
    Description:
    -----------
    Event triggerd when the value of the input field changes.
    A Date object containing the selected time is passed as the first argument of the callback.
    Note: the variable time is a function parameter received in the Javascript side

    Related Pages:

			https://api.jqueryui.com/autocomplete/#event-change

    :param jsFnc:
    """
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    self._jsStyles["change"] = "function( event, ui ){ %s }" % JsUtils.jsConvertFncs(jsFnc, toStr=True)
    return self

  def search(self, jsFnc, profile=None):
    """
    Description:
    -----------
    Triggered before a search is performed, after minLength and delay are met.
    If canceled, then no request will be started and no items suggested.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#event-search

    :param jsFnc:
    """
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    self._jsStyles["search"] = "function( event, ui ){ %s }" % JsUtils.jsConvertFncs(jsFnc, toStr=True)
    return self

  def focus(self, jsFnc, profile=None):
    """
    Description:
    -----------
    Triggered when focus is moved to an item (not selecting).
    The default action is to replace the text field's value with the value of the focused item, though only if the event was triggered by a keyboard interaction.
    Canceling this event prevents the value from being updated, but does not prevent the menu item from being focused.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#event-focus

    Attributes:
    ----------
    :param focus:
    :param profile:
    """
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    self._jsStyles["focus"] = "function( event, ui ){ %s }" % JsUtils.jsConvertFncs(jsFnc, toStr=True)
    return self

  def close(self, jsFnc, profile=None):
    """
    Description:
    -----------
    Triggered when the menu is hidden. Not every close event will be accompanied by a change event.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#event-close

    Attributes:
    ----------
    :param jsFnc:
    :param profile:
    """
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    self._jsStyles["close"] = "function( event, ui ){ %s }" % JsUtils.jsConvertFncs(jsFnc, toStr=True)
    return self

  def select(self, jsFnc, profile=None):
    """
    Description:
    -----------
    Triggered when an item is selected from the menu.
    The default action is to replace the text field's value with the value of the selected item.
    Canceling this event prevents the value from being updated, but does not prevent the menu from closing.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#event-select

    Attributes:
    ----------
    :param jsFnc:
    :param profile:
    """
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    self._jsStyles["select"] = "function( event, ui ){ %s }" % JsUtils.jsConvertFncs(jsFnc, toStr=True)
    return self

  def response(self, jsFnc, profile=None):
    """
    Description:
    -----------
    Triggered after a search completes, before the menu is shown.
    Useful for local manipulation of suggestion data, where a custom source option callback is not required.
    This event is always triggered when a search completes, even if the menu will not be shown because there are no results or the Autocomplete is disabled.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#event-response

    Attributes:
    ----------
    :param jsFnc:
    :param profile:
    """
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    self._jsStyles["response"] = "function( event, ui ){ %s }" % JsUtils.jsConvertFncs(jsFnc, toStr=True)


class FieldRange(Field):
  name = 'Field Range'

  def __init__(self, report, value, min, max, step, label, placeholder, icon, width, height, htmlCode, helper, options, profile):
    input = report.ui.inputs.d_range(value, min=min, max=max, step=step, width=(None, "%"), placeholder=placeholder, options=options)
    super(FieldRange, self).__init__(report, input, label, placeholder, icon, width, height, htmlCode, helper, profile)


class FieldCheckBox(Field):
  name = 'Field Checkbox'

  def __init__(self, report, value, label, icon, width, height, htmlCode, helper, options, profile):
    input = report.ui.inputs.checkbox(value, width=(None, "%"), options=options)
    super(FieldCheckBox, self).__init__(report, input, label, "", icon, width, height, htmlCode, helper, profile)
    self.style.css.line_height = Defaults.LINE_HEIGHT


class FieldInteger(Field):
  name = 'Field Integer'

  def __init__(self, report, value, label, placeholder, icon, width, height, htmlCode, helper, options, profile):
    input = report.ui.inputs.d_int(value, width=(None, "%"), placeholder=placeholder, options=options)
    super(FieldInteger, self).__init__(report, input, label, placeholder, icon, width, height, htmlCode, helper, profile)


class FieldFile(Field):
  name = 'Field Integer'

  def __init__(self, report, value, label, placeholder, icon, width, height, htmlCode, helper, options, profile):
    input = report.ui.inputs.file(value, width=(None, "%"), placeholder=placeholder, options=options)
    super(FieldFile, self).__init__(report, input, label, placeholder, icon, width, height, htmlCode, helper, profile)


class FieldPassword(Field):
  name = 'Field Password'

  def __init__(self, report, value, label, placeholder, icon, width, height, htmlCode, helper, options, profile):
    input = report.ui.inputs.password(value, width=(None, "%"), placeholder=placeholder, options=options)
    super(FieldPassword, self).__init__(report, input, label, placeholder, icon, width, height, htmlCode, helper, profile)


class FieldTextArea(Field):
  name = 'Field Textarea'

  def __init__(self, report, value, label, placeholder, icon, width, height, htmlCode, helper, options, profile):
    input = report.ui.inputs.textarea(value, width=(100, "%"), placeholder=placeholder, options=options)
    super(FieldTextArea, self).__init__(report, input, label, placeholder, icon, width, height, htmlCode, helper, profile)


class FieldSelect(Field):
  name = 'Field Select'

  def __init__(self, report, value, label, icon, width, height, htmlCode, helper, options, profile):
    input = report.ui.select(value, width=(100, "%"), options=options)
    super(FieldSelect, self).__init__(report, input, label, "", icon, width, height, htmlCode, helper, profile)


class Checkbox(Html.Html):
  name = 'Checkbox'

  def __init__(self, report, flag, label, group_name, width, height, htmlCode, options, attrs, profile):
    super(Checkbox, self).__init__(report, {"value": flag}, htmlCode=htmlCode, css_attrs={"width": width, "height": height},
                                   profile=profile, options=options)
    self.set_attrs(attrs={"type": "checkbox"})
    self.set_attrs(attrs=attrs)
    self.css({"cursor": 'pointer', 'display': 'inline-block', 'vertical-align': 'middle', 'margin-left': '2px'})
    self.style.css.line_height = Defaults.LINE_HEIGHT
    self._label = label or ''

  @property
  def dom(self):
    """
    Description:
    -----------

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsHtmlField.Check
    """
    if self._dom is None:
      self._dom = JsHtmlField.Check(self, report=self._report)
    return self._dom

  @property
  def _js__builder__(self):
    return '''htmlObj.checked = data.value; 
      if(data.text !== null){
        htmlObj.parentNode.insertBefore(document.createTextNode(data.text), htmlObj.nextSibling)};
      if(typeof options.css !== 'undefined'){for(var k in options.css){htmlObj.style[k] = options.css[k]}}'''

  def __str__(self):
    return '<input %(strAttr)s>%(label)s' % {'strAttr': self.get_attrs(pyClassNames=self.style.get_classes()), 'label': self._label}


class Radio(Html.Html):
  name = 'Radio'

  def __init__(self, report, flag, label, group_name, icon, width, height, htmlCode, helper, options, profile):
    super(Radio, self).__init__(report, {"value": flag, 'text': label}, htmlCode=htmlCode,
                                         css_attrs={"width": width, 'height': height}, profile=profile)
    self.add_input("", position="before", css={"width": 'none', "vertical-align": 'middle'})
    self.add_label(label, position="after", css={"display": 'inline-block', "width": "None", 'float': 'none'})
    self.input.set_attrs(name="data-content", value=label)
    if flag:
      self.input.set_attrs({"checked": json.dumps(flag)})
    self.input.style.clear()
    if group_name is not None:
      self.input.set_attrs(name="name", value=group_name)
    else:
      self.input.set_attrs(name="name", value=self.htmlCode)
    self.input.set_attrs(attrs={"type": "radio"})
    self.add_helper(helper, css={"line-height": '%spx' % Defaults.LINE_HEIGHT})
    self.input.css({"cursor": 'pointer', 'display': 'inline-block', 'vertical-align': 'middle', 'min-width': 'none'})
    self.css({'vertical-align': 'middle', 'text-align': "left"})
    self.add_icon(icon, position="after", css={"margin-left": '5px', 'color': self._report.theme.success[1]})

  @property
  def dom(self):
    """
    Description:
    -----------

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsHtmlField.Radio
    """
    if self._dom is None:
      self._dom = JsHtmlField.Radio(self, report=self._report)
    return self._dom

  @property
  def js(self):
    """
    Description:
    -----------
    Javascript Functions

    :rtype: JsComponents.Radio
    """
    if self._js is None:
      self._js = JsComponents.Radio(self, report=self._report)
    return self._js

  @property
  def _js__builder__(self):
    return '''htmlObj.checked = data.value; 
      if(data.text !== null){
        htmlObj.parentNode.insertBefore(document.createTextNode(data.text), htmlObj.nextSibling)};
      if(typeof options.css !== 'undefined'){for(var k in options.css){htmlObj.style[k] = options.css[k]}}'''

  def __str__(self):
    return '<div %(strAttr)s>%(helper)s</div>' % {'strAttr': self.get_attrs(pyClassNames=self.style.get_classes()), 'helper': self.helper}


class TextArea(Html.Html):
  name = 'Text Area'

  def __init__(self, report, text, width, rows, placeholder, background_color, htmlCode, options, profile):
    super(TextArea, self).__init__(report, text, htmlCode=htmlCode, css_attrs={"width": width, 'box-sizing': 'border-box'}, profile=profile)
    self.rows, self.backgroundColor = rows, background_color
    self.style.add_classes.input.textarea()
    self.set_attrs({"rows": rows, "placeholder": placeholder or ""})
    self.__options = OptInputs.OptionsTextarea(self, options)

  @property
  def options(self):
    """
    Description:
    -----------
    Property to set all the input component properties

    :rtype: OptInputs.OptionsTextarea
    """
    return self.__options

  def selectable(self, jsFncs=None, profile=False):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFncs:
    :param profile:

    :return: self. to allow the function chaining
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self.attr['onclick'] = "this.blur();this.select();%s" % JsUtils.jsConvertFncs(jsFncs, toStr=True)
    return self

  @property
  def dom(self):
    """
    Description:
    -----------
    Return the HTML DOM object

    Related Pages:

			https://www.w3schools.com/js/js_htmldom.asp

    :rtype: JsHtmlField.JsHtmlTextarea
    """
    if self._dom is None:
      self._dom = JsHtmlField.Textarea(self, report=self._report)
    return self._dom

  @property
  def _js__builder__(self):
    return 'htmlObj.innerHTML = data'

  def __str__(self):
    return '<textarea %(strAttr)s>%(val)s</textarea>' % {"strAttr": self.get_attrs(pyClassNames=self.style.get_classes()), 'val': self.val}


class Search(Html.Html):
  name = 'Search'

  def __init__(self, report, text, placeholder, color, height, htmlCode, tooltip, extensible, profile):
    super(Search, self).__init__(report, "", htmlCode=htmlCode, css_attrs={"height": height}, profile=profile)
    self.color = self._report.theme.colors[-1] if color is None else color
    self.css({"display": "inline-block", "margin-bottom": '2px', 'box-sizing': 'border-box'})
    #
    if not extensible:
      self.style.add_classes.layout.search()
      self.style.css.width = "100%"
    else:
      self.style.add_classes.layout.search_extension()
    self.add_input(text).input.set_attrs({"placeholder": placeholder, "spellcheck": False})
    self.input.css({"text-align": 'left', 'padding-left': '%spx' % Defaults.LINE_HEIGHT})
    self.add_icon("fas fa-search").icon.attr['id'] = "%s_button" % self.htmlCode
    self.icon.css({"margin": '6px 0 6px 5px', 'display': 'block', 'cursor': 'pointer', 'position': 'absolute'})
    self.tooltip(tooltip)

  @property
  def dom(self):
    if self._dom is None:
      self._dom = JsHtmlField.JsHtmlFields(self, report=self._report)
    return self._dom

  @property
  def _js__builder__(self):
    return '''htmlObj.find('input').val(data)'''

  def click(self, jsFncs, profile=False, source_event=None):
    return self.icon.click(jsFncs, profile, source_event)

  def enter(self, jsFncs, profile=False):
    """
    Description:
    -----------
    Add an javascript action when the key enter is pressed on the keyboard

    Example
    htmlObj.enter(" alert() ")

    Attributes:
    ----------
    :param jsFncs:

    :return: The python object itself
    """
    self.click(jsFncs)
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    return self.on("keydown", ["if (event.keyCode  == 13) {event.preventDefault(); %(jsFnc)s } " % {"jsFnc": JsUtils.jsConvertFncs(jsFncs, toStr=True)}], profile=profile)

  def __str__(self):
    return '<div %(attr)s></div>' % {"attr": self.get_attrs(pyClassNames=self.style.get_classes())}




from epyk.core.py import PyCrypto


#import doctest
#doctest.testmod()
message = "http://192.168.0.34:5000/reports/run/examples/TestSlider?slider_1788657270680=[FormatDate($(%27&slider_2122911917784=[FormatDate($(%27&slider_1662408041416=[FormatDate($(%27&slider_1526638297440=[FormatDate($(%27&slider_2905684505600=[FormatDate($(%27#"
crt = PyCrypto.PyCrypto()


encrypted = crt.encrypt(message)
print(encrypted)
print(len(encrypted))
decrypted = crt.decrypt(encrypted)
print(decrypted)